#!/usr/bin/env python
# -*- coding: utf-8 -*-
import logging
from typing import Union
from dask.array.core import Array as DaskArray
from numpy import ndarray as NumpyArray
from importlib import import_module
from pathlib import Path
import tempfile
import shutil
import numpy as np
from bioio import BioImage
from bioio.writers import OmeTiffWriter
import torch
from mmv_im2im.utils.misc import generate_test_dataset_dict, parse_config
from mmv_im2im.utils.for_transform import parse_monai_ops_vanilla
from skimage.io import imsave as save_rgb
import bioio_tifffile
from tqdm import tqdm
from monai.inferers import sliding_window_inference


# https://pytorch-lightning.readthedocs.io/en/latest/starter/introduction_guide.html#predicting
###############################################################################

log = logging.getLogger(__name__)

###############################################################################


class ProjectTester(object):
    """
    entry for training models

    Parameters
    ----------
    cfg: configuration
    """

    def __init__(self, cfg):
        # extract the three major chuck of the config
        self.model_cfg = cfg.model
        self.data_cfg = cfg.data

        # define variables
        self.model = None
        self.data = None
        self.pre_process = None
        self.cpu = False
        self.spatial_dims = -1

    def setup_model(self):
        model_category = self.model_cfg.framework
        model_module = import_module(f"mmv_im2im.models.pl_{model_category}")
        my_model_func = getattr(model_module, "Model")
        self.model = my_model_func(self.model_cfg, train=False)

        if (
            self.model_cfg.model_extra is not None
            and "cpu_only" in self.model_cfg.model_extra
            and self.model_cfg.model_extra["cpu_only"]
        ):
            self.cpu = True
            checkpoint = torch.load(
                self.model_cfg.checkpoint,
                map_location=torch.device("cpu"),
                weights_only=False,
            )
        else:
            checkpoint = torch.load(self.model_cfg.checkpoint, weights_only=False)

        if isinstance(checkpoint, dict) and "state_dict" in checkpoint:

            pre_train = checkpoint
            pre_train["state_dict"].pop("criterion.xym", None)
            pre_train["state_dict"].pop("criterion.xyzm", None)
            self.model.load_state_dict(pre_train["state_dict"], strict=False)
        else:

            state_dict = checkpoint
            state_dict.pop("criterion.xym", None)
            state_dict.pop("criterion.xyzm", None)
            self.model.load_state_dict(state_dict, strict=False)

        if not self.cpu:
            self.model.cuda()

        self.model.eval()

    def setup_data_processing(self):
        # determine spatial dimension from reader parameters
        if "Z" in self.data_cfg.inference_input.reader_params["dimension_order_out"]:
            self.spatial_dims = 3
        else:
            self.spatial_dims = 2

        # prepare data preprocessing if needed
        if self.data_cfg.preprocess is not None:
            # load preprocessing transformation
            self.pre_process = parse_monai_ops_vanilla(self.data_cfg.preprocess)

    def process_one_image(
        self, img: Union[DaskArray, NumpyArray], out_fn: Union[str, Path] = None
    ):

        if isinstance(img, DaskArray):
            # Perform the prediction
            x = img.compute()

        elif isinstance(img, NumpyArray):
            x = img
        else:
            raise ValueError("invalid image")

        # check if need to add channel dimension
        if len(x.shape) == self.spatial_dims:
            x = np.expand_dims(x, axis=0)

        # convert the numpy array to float tensor
        x = torch.tensor(x.astype(np.float32))

        # run pre-processing on tensor if needed

        if self.pre_process is not None:
            x = self.pre_process(x)

        # choose different inference function for different types of models
        # the input here is assumed to be a tensor
        with torch.no_grad():
            # add batch dimension and move to GPU

            if self.cpu:
                x = torch.unsqueeze(x, dim=0)
            else:
                x = torch.unsqueeze(x, dim=0).cuda()

            # TODO: add convert to tensor with proper type, similar to torchio check

            if (
                self.model_cfg.model_extra is not None
                and "sliding_window_params" in self.model_cfg.model_extra
            ):
                y_hat = sliding_window_inference(
                    inputs=x,
                    predictor=self.model,
                    device=torch.device("cpu"),
                    **self.model_cfg.model_extra["sliding_window_params"],
                )

                # currently, we keep sliding window stiching step on CPU, but assume
                # the output is on GPU (see note below). So, we manually move the data
                # back to GPU
                if not self.cpu:
                    y_hat = y_hat.cuda()
            else:
                y_hat = self.model(x)

        ###############################################################################
        #
        # Note: currently, we assume y_hat is still on gpu, because embedseg clustering
        # step is still only running on GPU (possible on CPU, need to some update on
        # grid loading). All the post-procesisng functions we tested so far can accept
        # tensor on GPU. If it is from mmv_im2im.post_processing, it will automatically
        # convert the tensor to a numpy array and return the result as numpy array; if
        # it is from monai.transforms, it is tensor in and tensor out. We have two items
        # as #TODO: (1) we will extend post-processing functions in mmv_im2im to work
        # similarly to monai transforms, ie. ndarray in ndarray out or tensor in tensor
        # out. (2) allow yaml config to control if we want to run post-processing on
        # GPU tensors or ndarrays
        #
        ##############################################################################

        # do post-processing on the prediction
        if self.data_cfg.postprocess is not None:
            pp_data = y_hat
            for pp_info in self.data_cfg.postprocess:
                pp = parse_config(pp_info)
                pp_data = pp(pp_data)
            if torch.is_tensor(pp_data):
                pred = pp_data.cpu().numpy()
            else:
                pred = pp_data
        else:
            pred = y_hat.cpu().numpy()

        if out_fn is None:
            return pred

        # determine output dimension orders
        if out_fn.suffix == ".npy":
            np.save(out_fn, pred)
        else:
            if len(pred.shape) == 2:
                OmeTiffWriter.save(pred, out_fn, dim_order="YX")
            elif len(pred.shape) == 3:
                # 3D output, for 2D data
                if self.spatial_dims == 2:
                    # save as RGB or multi-channel 2D
                    if pred.shape[0] == 3:
                        if out_fn.suffix != ".png":
                            out_fn = out_fn.with_suffix(".png")
                        save_rgb(out_fn, np.moveaxis(pred, 0, -1))
                    else:
                        OmeTiffWriter.save(pred, out_fn, dim_order="CYX")
                elif self.spatial_dims == 3:
                    OmeTiffWriter.save(pred, out_fn, dim_order="ZYX")
                else:
                    raise ValueError("Invalid spatial dimension of pred")
            elif len(pred.shape) == 4:
                if self.spatial_dims == 3:
                    OmeTiffWriter.save(pred, out_fn, dim_order="CZYX")
                elif self.spatial_dims == 2:
                    if pred.shape[0] == 1:
                        if pred.shape[1] == 1:
                            OmeTiffWriter.save(pred[0, 0], out_fn, dim_order="YX")
                        elif pred.shape[1] == 3:
                            if out_fn.suffix != ".png":
                                out_fn = out_fn.with_suffix(".png")
                            save_rgb(
                                out_fn,
                                np.moveaxis(
                                    pred[0,],
                                    0,
                                    -1,
                                ),
                            )
                        else:
                            OmeTiffWriter.save(
                                pred[0,],
                                out_fn,
                                dim_order="CYX",
                            )
                    else:
                        raise ValueError("invalid 4D output for 2d data")
            elif len(pred.shape) == 5:
                assert pred.shape[0] == 1, "error, found non-trivial batch dimension"
                OmeTiffWriter.save(
                    pred[0,],
                    out_fn,
                    dim_order="CZYX",
                )
            else:
                raise ValueError("error in prediction output shape")

    def run_inference(self):

        self.setup_model()
        self.setup_data_processing()
        # set up data filenames
        dataset_list = generate_test_dataset_dict(
            self.data_cfg.inference_input.dir, self.data_cfg.inference_input.data_type
        )

        # loop through all images and apply the model
        for ds in tqdm(dataset_list, desc="Predicting images"):

            # output file name info
            fn_core = Path(ds).stem
            suffix = self.data_cfg.inference_output.suffix

            timelapse_data = 0
            # if timelapse ...
            if (
                "T"
                in self.data_cfg.inference_input.reader_params["dimension_order_out"]
            ):
                if "T" in self.data_cfg.inference_input.reader_params:
                    raise NotImplementedError(
                        "processing a subset of all timepoint is not supported yet"
                    )
                tmppath = tempfile.mkdtemp()
                print(f"making a temp folder at {tmppath}")

                # get the number of time points
                try:
                    reader = BioImage(ds, reader=bioio_tifffile.Reader)
                except Exception:
                    try:
                        reader = BioImage(ds)
                    except Exception as e:
                        print(f"Error: {e}")
                        print(f"Image {ds} failed at read process check the format.")

                timelapse_data = reader.dims.T

                tmpfile_list = []
                for t_idx in tqdm(
                    range(timelapse_data), desc="Predicting image timepoint"
                ):
                    try:
                        img = BioImage(ds, reader=bioio_tifffile.Reader).get_image_data(
                            T=[t_idx], **self.data_cfg.inference_input.reader_params
                        )
                    except Exception:
                        try:
                            img = BioImage(ds).get_image_data(
                                T=[t_idx], **self.data_cfg.inference_input.reader_params
                            )
                        except Exception as e:
                            print(f"Error: {e}")
                            print(
                                f"Image {ds} failed at read process check the format."
                            )

                    # prepare output filename
                    out_fn = Path(tmppath) / f"{fn_core}_{t_idx}.npy"

                    self.process_one_image(img, out_fn)
                    tmpfile_list.append(out_fn)

                # gather all individual outputs and save as timelapse
                out_array = [np.load(tmp_one_file) for tmp_one_file in tmpfile_list]
                out_array = np.stack(out_array, axis=0)

                # prepare output filename
                if "." in suffix:
                    if ".tif" in suffix or ".tiff" in suffix or ".ome.tif" in suffix:
                        out_fn = (
                            Path(self.data_cfg.inference_output.path)
                            / f"{fn_core}{suffix}"
                        )
                    else:
                        raise ValueError(
                            "please check output suffix, either unexpected dot or unsupported fileformat"  # noqa E501
                        )
                else:
                    out_fn = (
                        Path(self.data_cfg.inference_output.path)
                        / f"{fn_core}{suffix}.tiff"
                    )

                if len(out_array.shape) == 3:
                    dim_order = "TYX"
                elif len(out_array.shape) == 4:
                    if self.spatial_dims == 3:
                        dim_order = "TZYX"
                    else:
                        dim_order = "TCYX"
                elif len(out_array.shape) == 5:
                    dim_order = "TCZYX"
                else:
                    raise ValueError(f"Unexpected pred of shape {out_array.shape}")

                # save the file output
                OmeTiffWriter.save(out_array, out_fn, dim_order=dim_order)

                # clean up temporary dir
                shutil.rmtree(tmppath)
            else:
                try:
                    img = BioImage(ds, reader=bioio_tifffile.Reader).get_image_data(
                        **self.data_cfg.inference_input.reader_params
                    )
                except Exception:
                    try:
                        img = BioImage(ds).get_image_data(
                            **self.data_cfg.inference_input.reader_params
                        )
                    except Exception as e:
                        print(f"Error: {e}")
                        print(f"Image {ds} failed at read process check the format.")

                # prepare output filename
                if "." in suffix:
                    if (
                        ".png" in suffix
                        or ".tif" in suffix
                        or ".tiff" in suffix
                        or ".ome.tif" in suffix
                    ):
                        out_fn = (
                            Path(self.data_cfg.inference_output.path)
                            / f"{fn_core}{suffix}"
                        )
                    else:
                        raise ValueError(
                            "please check output suffix, either unexpected dot or unsupported fileformat"  # noqa E501
                        )
                else:
                    out_fn = (
                        Path(self.data_cfg.inference_output.path)
                        / f"{fn_core}{suffix}.tiff"
                    )

                self.process_one_image(img, out_fn)



import torch
import torch.nn as nn
import torch.nn.functional as F
from mmv_im2im.utils.fractal_layers import FractalDimension
from mmv_im2im.utils.topological_loss import TI_Loss
from mmv_im2im.utils.connectivity_loss import ConnectivityCoherenceLoss
from monai.losses import GeneralizedDiceFocalLoss
from monai.metrics import HausdorffDistanceMetric


class KLDivergence(nn.Module):
    """Calculates KL Divergence between two diagonal Gaussians."""

    def __init__(self):
        super().__init__()

    def forward(self, mu_q, logvar_q, mu_p, logvar_p, kl_clamp=None):
        """
        Calculates the KL Divergence between two diagonal Gaussian distributions.

        Args:
            mu_q (torch.Tensor): Mean of the approximate posterior distribution.
            logvar_q (torch.Tensor): Log-variance of the approximate posterior
                                     distribution.
            mu_p (torch.Tensor): Mean of the prior distribution.
            logvar_p (torch.Tensor): Log-variance of the prior distribution.
            clamp (float): Value to clamp logvar_q, logvar_p in case of gradient explotion.

        Returns:
            torch.Tensor: The mean KL divergence over the batch.
        """
        # Clamp log-variances to prevent numerical instability
        # This limits exp(logvar) to a stable range, e.g., [2.06e-9, 4.85e8]
        if kl_clamp is not None:
            logvar_q = torch.clamp(logvar_q, min=-kl_clamp, max=kl_clamp)
            logvar_p = torch.clamp(logvar_p, min=-kl_clamp, max=kl_clamp)

        kl_batch_sum = 0.5 * torch.sum(
            logvar_p
            - logvar_q
            + (torch.exp(logvar_q) + (mu_q - mu_p) ** 2) / torch.exp(logvar_p)
            - 1,
            dim=[1, 2, 3],  # Sum over latent channels, H, W
        )
        return torch.mean(kl_batch_sum)


class ELBOLoss(nn.Module):
    """
    Calculates the Evidence Lower Bound (ELBO) loss for Probabilistic UNet,
    with optional fractal dimension, topological, and connectivity regularization.

    Args:
        beta (float): Weighting factor for the KL divergence term.
        n_classes (int): Number of classes in the segmentation task.
        kl_clamp (float): Value to clamp logvar_q, logvar_p in case of gradient explotion for kl.
        use_fractal_regularization (bool): If True, includes the fractal dimension regularization term.
        fractal_weight (float): Weighting factor for the fractal dimension loss term (only if use_fractal_regularization is True).
        fractal_num_kernels (int): Number of kernels for FractalDimension (only if use_fractal_regularization is True).
        fractal_mode (str): Mode for FractalDimension ("classic" or "entropy") (only if use_fractal_regularization is True).
        fractal_to_binary (bool): Whether to binarize input for FractalDimension (only if use_fractal_regularization is True).
        use_topological_regularization (bool): If True, includes the topological regularization term.
        topological_weight (float): Weighting factor for the topological loss term (only if use_topological_regularization is True).
        topological_dim (int): Dimension for TI_Loss (2 for 2D, 3 for 3D) (only if use_topological_regularization is True).
        topological_connectivity (int): Connectivity for TI_Loss (4 or 8 for 2D; 6 or 26 for 3D) (only if use_topological_regularization is True).
        topological_inclusion (list): List of [A,B] class pairs for inclusion in TI_Loss (only if use_topological_regularization is True).
        topological_exclusion (list): List of [A,C] class pairs for exclusion in TI_Loss (only if use_topological_regularization is True).
        topological_min_thick (int): Minimum thickness for TI_Loss (only if use_topological_regularization is True and connectivity is 8 or 26).
        use_connectivity_regularization (bool): If True, includes the new connectivity coherence regularization term.
        connectivity_weight (float): Weighting factor for the connectivity coherence loss term.
        connectivity_kernel_size (int): Kernel size for connectivity coherence loss (e.g., 3).
        connectivity_ignore_background (bool): If True, ignore background for connectivity loss.
        elbo_class_weights (list or torch.Tensor, optional): Weights for each class in the cross-entropy loss.
        use_gdl_focal_regularization (bool): If True, Includes Generalized Dice Focal (GDF) regularization.
        gdl_focal_weight (float): Weighting factor for GDF.
        gdl_class_weights (list):  Weights for each class.
        use_hausdorff_regularization (bool): Enable or disable  Hausdorff regularization
        hausdorff_weight (float): Weighting factor for the Hausdorff loss term.
        hausdorff_ignore_background (bool): If True, ignore background for Hausdorff loss.
        task (str): The task type, either "segment" for segmentation or "regression" for regression.
    """

    def __init__(
        self,
        beta: float = 1.0,
        n_classes: int = 2,
        kl_clamp: float = None,
        use_fractal_regularization: bool = False,
        fractal_weight: float = 0.1,
        fractal_num_kernels: int = 5,
        fractal_mode: str = "classic",
        fractal_to_binary: bool = True,
        use_topological_regularization: bool = False,
        topological_weight: float = 0.1,
        topological_dim: int = 2,
        topological_connectivity: int = 4,
        topological_inclusion: list = None,
        topological_exclusion: list = None,
        topological_min_thick: int = 1,
        use_connectivity_regularization: bool = False,
        connectivity_weight: float = 0.1,
        connectivity_kernel_size: int = 3,
        connectivity_ignore_background: bool = True,
        use_gdl_focal_regularization: bool = False,
        gdl_focal_weight: float = 1.0,
        elbo_class_weights: list = None,
        gdl_class_weights: list = None,
        use_hausdorff_regularization: bool = False,
        hausdorff_weight: float = 0.1,
        hausdorff_ignore_background: bool = True,
        task: str = "segment",
    ):
        super().__init__()
        self.beta = beta
        self.n_classes = n_classes
        self.kl_clamp = kl_clamp
        self.task = task
        self.kl_divergence_calculator = KLDivergence()

        self.use_fractal_regularization = use_fractal_regularization
        if self.use_fractal_regularization:
            self.fractal_weight = fractal_weight
            self.fractal_dimension_calculator = FractalDimension(
                num_kernels=fractal_num_kernels,
                mode=fractal_mode,
                to_binary=fractal_to_binary,
            )
        else:
            self.fractal_weight = 0.0

        self.use_topological_regularization = use_topological_regularization
        if self.use_topological_regularization:
            self.topological_weight = topological_weight
            if topological_inclusion is None:
                topological_inclusion = []
            if topological_exclusion is None:
                topological_exclusion = []
            self.topological_loss_calculator = TI_Loss(
                dim=topological_dim,
                connectivity=topological_connectivity,
                inclusion=topological_inclusion,
                exclusion=topological_exclusion,
                min_thick=topological_min_thick,
            )
        else:
            self.topological_weight = 0.0

        self.use_connectivity_regularization = use_connectivity_regularization
        if self.use_connectivity_regularization:
            self.connectivity_weight = connectivity_weight
            self.connectivity_coherence_calculator = ConnectivityCoherenceLoss(
                kernel_size=connectivity_kernel_size,
                ignore_background=connectivity_ignore_background,
                num_classes=n_classes,
            )
        else:
            self.connectivity_weight = 0.0

        self.use_gdl_focal_regularization = use_gdl_focal_regularization
        if self.use_gdl_focal_regularization:
            self.gdl_focal_weight = gdl_focal_weight
            monai_focal_weights = None
            if gdl_class_weights is not None:
                monai_focal_weights = torch.tensor(
                    gdl_class_weights, dtype=torch.float32
                )
            self.gdl_focal_loss_calculator = GeneralizedDiceFocalLoss(
                softmax=True, to_onehot_y=True, weight=monai_focal_weights
            )
        else:
            self.gdl_focal_weight = 0.0

        self.use_hausdorff_regularization = use_hausdorff_regularization
        if self.use_hausdorff_regularization:
            self.hausdorff_weight = hausdorff_weight
            self.hausdorff_distance_calculator = HausdorffDistanceMetric(
                include_background=not hausdorff_ignore_background,
                reduction="mean",
            )
        else:
            self.hausdorff_weight = 0.0

        if elbo_class_weights is not None:
            self.elbo_class_weights = torch.tensor(
                elbo_class_weights, dtype=torch.float32
            )
        else:
            self.elbo_class_weights = None

    def forward(self, logits, y_true, prior_mu, prior_logvar, post_mu, post_logvar):
        """
        Computes the ELBO loss, with optional fractal dimension and topological regularization terms.

        Args:
            logits (torch.Tensor): Output logits from the Probabilistic UNet
                                   (B, C, H, W).
            y_true (torch.Tensor): Ground truth segmentation mask (B, 1, H, W
                                   or B, H, W).
            prior_mu (torch.Tensor): Mean of the prior distribution.
            prior_logvar (torch.Tensor): Log-variance of the prior distribution.
            post_mu (torch.Tensor): Mean of the approximate posterior distribution.
            post_logvar (torch.Tensor): Log-variance of the approximate posterior
                                        distribution.

        Returns:
            torch.Tensor: The calculated ELBO loss.
        """

        if self.task == "segment":

            if y_true.shape[1] == 1:
                y_true_squeezed = y_true.squeeze(1)
            else:
                y_true_squeezed = y_true

            if (
                self.elbo_class_weights is not None
                and self.elbo_class_weights.device != logits.device
            ):
                elbo_class_weights_on_device = self.elbo_class_weights.to(logits.device)
            else:
                elbo_class_weights_on_device = self.elbo_class_weights

            log_likelihood = -F.cross_entropy(
                logits,
                y_true_squeezed.long(),
                reduction="mean",
                weight=elbo_class_weights_on_device,
            )

        elif self.task == "regression":
            # log_likelihood = -F.mse_loss(
            #     logits,
            #     y_true,
            #     reduction="mean",
            # )
            # log_likelihood = -F.mae_loss(
            #     logits,
            #     y_true,
            #     reduction="mean",
            # )
            log_likelihood = -F.huber_loss(
                logits,
                y_true,
                reduction="mean",
            )
        else:
            raise ValueError(f"Unknown task type: {self.task}")

        # KL-Divergence
        kl_div = self.kl_divergence_calculator(
            post_mu, post_logvar, prior_mu, prior_logvar, self.kl_clamp
        )

        elbo_loss = -(log_likelihood - self.beta * kl_div)

        total_loss = elbo_loss

        if self.task == "segment":
            if self.use_fractal_regularization:
                y_pred_mask = (
                    F.softmax(logits, dim=1).argmax(dim=1, keepdim=True).float()
                )

                if y_true.ndim == 4 and y_true.shape[1] == 1:
                    y_true_for_fractal = y_true.float()
                else:
                    y_true_for_fractal = y_true.unsqueeze(1).float()

                fd_true = self.fractal_dimension_calculator(y_true_for_fractal)
                fd_pred = self.fractal_dimension_calculator(y_pred_mask)

                fractal_loss = torch.mean(torch.abs(fd_true - fd_pred))
                total_loss += self.fractal_weight * fractal_loss

            if self.use_topological_regularization:
                # y_true needs to be B, C, H, W or B, C, H, W, D for TI_Loss, where C=1
                # If y_true is B, H, W, unsqueeze to B, 1, H, W
                if y_true.ndim == 3:
                    y_true_for_topological = y_true.unsqueeze(1).float()
                else:
                    y_true_for_topological = y_true.float()

                # logits are B, C, H, W (or B, C, H, W, D), which is what TI_Loss expects for x
                topological_loss = self.topological_loss_calculator(
                    logits, y_true_for_topological
                )
                total_loss += self.topological_weight * topological_loss

            if self.use_connectivity_regularization:
                # y_pred_softmax: (B, C, H, W)
                y_pred_softmax = F.softmax(logits, dim=1)

                # y_true_one_hot: Need to convert y_true_squeezed (B, H, W) to one-hot (B, C, H, W)
                y_true_one_hot = (
                    F.one_hot(y_true_squeezed.long(), num_classes=self.n_classes)
                    .permute(0, 3, 1, 2)
                    .float()
                )

                connectivity_loss = self.connectivity_coherence_calculator(
                    y_pred_softmax, y_true_one_hot
                )
                total_loss += self.connectivity_weight * connectivity_loss

            if self.use_gdl_focal_regularization:
                # logits: (B, C, H, W)
                # y_true: (B, H, W) o (B, 1, H, W)
                y_true_for_gdl_focal = y_true_squeezed.unsqueeze(1).long()
                gdl_focal_loss = self.gdl_focal_loss_calculator(
                    logits, y_true_for_gdl_focal
                )
                total_loss += self.gdl_focal_weight * gdl_focal_loss

            if self.use_hausdorff_regularization:
                # Get the one-hot encoded ground truth
                # Squeeze y_true to (B, H, W) if it's (B, 1, H, W)
                if y_true.ndim == 4 and y_true.shape[1] == 1:
                    y_true_squeezed = y_true.squeeze(1)
                else:
                    y_true_squeezed = y_true

                # Calculate the Hausdorff distance
                # The metric returns a tensor of shape (B, C), so we take the mean
                try:
                    # Convert ground truth to one-hot format (B, C, H, W)
                    y_true_one_hot = F.one_hot(
                        y_true_squeezed.long(), num_classes=self.n_classes
                    ).permute(0, 3, 1, 2)

                    # Get the one-hot encoded prediction from logits
                    y_pred_one_hot = F.one_hot(
                        logits.argmax(dim=1), num_classes=self.n_classes
                    ).permute(0, 3, 1, 2)

                    # Calculate the Hausdorff distance
                    # The metric returns a tensor of shape (B, C), so we take the mean
                    hausdorff_loss = self.hausdorff_distance_calculator(
                        y_pred=y_pred_one_hot, y_true=y_true_one_hot
                    ).mean()

                    # Add the Hausdorff loss to the total loss
                    total_loss += self.hausdorff_weight * hausdorff_loss
                except Exception:
                    # Avoid troubles with some tensor that become None douring the hausdorff computation
                    pass

        return total_loss



import torch
import torch.nn as nn
import torch.nn.functional as F


class ConnectivityCoherenceLoss(nn.Module):
    """
    Calculates a connectivity coherence loss to penalize fragmentation or
    undesired isolated components within predicted regions.

    This loss encourages predicted regions to be spatially coherent and continuous
    by comparing local neighborhoods in predictions with ground truth.
    It penalizes:
    1. Discontinuities within what should be a single, connected region (e.g., breaking a vein).
    2. Isolated 'islands' of one class within another class.

    Args:
        kernel_size (int): Size of the convolutional kernel for neighborhood analysis (e.g., 3 for 3x3).
        ignore_background (bool): If True, the loss focuses primarily on non-background classes.
                                  Useful if background fragmentation is less critical.
        num_classes (int): Total number of classes, including background.
    """

    def __init__(
        self, kernel_size: int = 3, ignore_background: bool = True, num_classes: int = 2
    ):
        super().__init__()
        if kernel_size % 2 == 0:
            raise ValueError("kernel_size must be an odd number.")
        self.kernel_size = kernel_size
        self.ignore_background = ignore_background
        self.num_classes = num_classes
        self.average_kernel = torch.ones(1, 1, kernel_size, kernel_size) / (
            kernel_size**2 - 1
        )
        self.center_offset = (kernel_size // 2, kernel_size // 2)

    def forward(self, y_pred_softmax, y_true_one_hot):
        """
        Args:
            y_pred_softmax (torch.Tensor): Softmax probabilities from the model (B, C, H, W).
            y_true_one_hot (torch.Tensor): Ground truth as one-hot encoded tensor (B, C, H, W).
                                           Should be float.

        Returns:
            torch.Tensor: The calculated connectivity coherence loss.
        """
        current_average_kernel = self.average_kernel.to(y_pred_softmax.device)

        y_true_one_hot = y_true_one_hot.float()

        loss_per_class = []

        for c in range(self.num_classes):
            if self.ignore_background and c == 0:
                continue

            true_mask_c = y_true_one_hot[:, c : c + 1, :, :]
            pred_prob_c = y_pred_softmax[:, c : c + 1, :, :]

            padded_true_mask_c = F.pad(
                true_mask_c,
                (
                    self.center_offset[1],
                    self.center_offset[1],
                    self.center_offset[0],
                    self.center_offset[0],
                ),
                mode="replicate",
            )

            neighbor_sum_true = F.conv2d(
                padded_true_mask_c, current_average_kernel, padding=0, groups=1
            ) * (self.kernel_size**2)

            padded_pred_prob_c = F.pad(
                pred_prob_c,
                (
                    self.center_offset[1],
                    self.center_offset[1],
                    self.center_offset[0],
                    self.center_offset[0],
                ),
                mode="replicate",
            )

            pred_neighbor_avg = F.conv2d(
                padded_pred_prob_c, current_average_kernel, padding=0, groups=1
            )

            true_neighbor_avg = neighbor_sum_true / (self.kernel_size**2 - 1)

            loss_b = F.mse_loss(pred_neighbor_avg, true_mask_c, reduction="none")
            loss_c = F.mse_loss(pred_prob_c, true_neighbor_avg, reduction="none")

            class_coherence_loss = torch.mean(loss_b + loss_c)
            loss_per_class.append(class_coherence_loss)

        if not loss_per_class:
            return torch.tensor(0.0, device=y_pred_softmax.device)

        return torch.sum(torch.stack(loss_per_class))



import numpy as np
from bioio import BioImage
from bioio.writers import OmeTiffWriter
from tqdm import tqdm
import os
from typing import Callable


def Im2Fourier(image: np.ndarray, mode: str = "complex") -> np.ndarray:
    """
    Computes the 2D Fourier Transform for a multi-channel image and returns the result
    in different formats based on the specified mode.

    Args:
        image (np.ndarray): The input image as a NumPy array of shape (C, Y, X).
        mode (str): The desired output format. Must be one of "complex", "real", or "freq".
                    - "complex": Returns a matrix of shape (2C, Y, X) with the real and
                                 imaginary parts concatenated along the channel axis.
                    - "real": Returns a matrix of shape (C, Y, X) with only the real
                              part of the Fourier Transform.
                    - "freq": Returns a matrix of shape (2C, Y, X) with the magnitude
                              (abs) and phase (angle) for each channel.

    Returns:
        np.ndarray: The transformed image matrix according to the specified mode.
                    Returns None if the mode is invalid.
    """

    allowed_modes = ["complex", "real", "freq"]
    if mode not in allowed_modes:
        raise ValueError(f"Invalid mode: '{mode}'. Must be one of {allowed_modes}.")

    if image.ndim != 3:
        raise ValueError("Input image must be a 3D array of shape (C, Y, X).")

    C, Y, X = image.shape

    transformed_channels = []

    for c in range(C):

        channel = image[c, :, :]

        f_transform = np.fft.fft2(channel)

        f_transform_shifted = np.fft.fftshift(f_transform)

        if mode == "complex":
            real_part = np.real(f_transform_shifted)
            imag_part = np.imag(f_transform_shifted)
            transformed_channels.append(real_part)
            transformed_channels.append(imag_part)

        elif mode == "real":
            real_part = np.real(f_transform_shifted)
            transformed_channels.append(real_part)

        elif mode == "freq":
            magnitude = np.abs(f_transform_shifted)
            phase = np.angle(f_transform_shifted)
            transformed_channels.append(magnitude)
            transformed_channels.append(phase)

    return np.stack(transformed_channels, axis=0)


def Fourier2Im(fourier_image: np.ndarray, mode: str = "complex") -> np.ndarray:
    """
    Computes the 2D Inverse Fourier Transform for a multi-channel image and
    returns the result. It automatically deduces the number of original channels
    based on the input shape and mode.

    Args:
        fourier_image (np.ndarray): The input Fourier-transformed image. Its shape
                                    depends on the specified mode.
        mode (str): The format of the input. Must be one of "complex", "real", or "freq".

    Returns:
        np.ndarray: The reconstructed image matrix of shape (C, Y, X).
    """
    allowed_modes = ["complex", "real", "freq"]
    if mode not in allowed_modes:
        raise ValueError(f"Invalid mode: '{mode}'. Must be one of {allowed_modes}.")

    if fourier_image.ndim != 3:
        raise ValueError("Input image must be a 3D array of shape (N, Y, X).")

    N, Y, X = fourier_image.shape

    # Deduce the number of original channels based on the mode
    if mode == "complex":
        if N % 2 != 0:
            raise ValueError(
                f"For mode 'complex', the number of channels ({N}) must be even."
            )
        C = N // 2
    elif mode == "freq":
        if N % 2 != 0:
            raise ValueError(
                f"For mode 'freq', the number of channels ({N}) must be even."
            )
        C = N // 2
    else:  # mode == "real"
        C = N

    reconstructed_channels = []

    if mode == "complex":
        for i in range(C):
            real_part = fourier_image[2 * i, :, :]
            imag_part = fourier_image[2 * i + 1, :, :]
            f_transform_shifted = real_part + 1j * imag_part
            f_transform = np.fft.ifftshift(f_transform_shifted)
            reconstructed_channel = np.fft.ifft2(f_transform)
            reconstructed_channels.append(np.real(reconstructed_channel))

    elif mode == "real":
        for i in range(C):
            f_transform_shifted = fourier_image[i, :, :]
            f_transform = np.fft.ifftshift(f_transform_shifted)
            reconstructed_channel = np.fft.ifft2(f_transform)
            reconstructed_channels.append(np.real(reconstructed_channel))

    elif mode == "freq":
        for i in range(C):
            magnitude = fourier_image[2 * i, :, :]
            phase = fourier_image[2 * i + 1, :, :]
            f_transform_shifted = magnitude * np.exp(1j * phase)
            f_transform = np.fft.ifftshift(f_transform_shifted)
            reconstructed_channel = np.fft.ifft2(f_transform)
            reconstructed_channels.append(np.real(reconstructed_channel))

    return np.stack(reconstructed_channels, axis=0)


def discretizer2n(matrix: np.ndarray, n: int = 3) -> np.ndarray:
    """
    Quantizes the values of a float matrix to a specified number of integer ranges.

    The thresholds are calculated by dividing the range of matrix values
    into n equal parts.

    Args:
        matrix (np.ndarray): The input matrix with float values.
        n (int): The number of integer levels to quantize to (e.g., 3 for 0, 1, 2).

    Returns:
        np.ndarray: A new matrix with the quantized integer values (from 0 to n-1).
    """
    if not isinstance(matrix, np.ndarray):
        raise TypeError("Input must be a NumPy array.")
    if not isinstance(n, int) or n < 2:
        raise ValueError("n_levels must be an integer greater than or equal to 2.")

    if matrix.size == 0:
        return np.array([], dtype=int)

    min_val = np.min(matrix)
    max_val = np.max(matrix)

    if min_val == max_val:
        return np.full_like(matrix, int(n / 2), dtype=int)

    data_range = max_val - min_val
    step = data_range / n

    thresholds = [min_val + i * step for i in range(1, n)]

    quantized_matrix = np.zeros_like(matrix, dtype=int)

    for i in range(n - 1):
        quantized_matrix[matrix >= thresholds[i]] = i + 1

    return quantized_matrix.astype(np.uint8)


def Vol2Fourier(volume: np.ndarray, mode: str = "complex") -> np.ndarray:
    """
    Applies the 2D Fourier Transform to each image (slice) of a volume and
    returns a new volume of transformations.

    Args:
        volume (np.ndarray): The input volume as a 4D NumPy array
                             with shape (Z, C, Y, X).
        mode (str): The desired output format. Must be one of "complex",
                    "real", or "freq".

    Returns:
        np.ndarray: The transformed volume with shape (Z, C', Y, X), where C'
                    is the number of channels of the transformation.
    """
    if volume.ndim != 4:
        raise ValueError("Input volume must be a 4D array with shape (Z, C, Y, X).")

    Z, C, Y, X = volume.shape
    transformed_slices = []

    for z in range(Z):
        image_slice = volume[z, :, :, :]
        transformed_slice = Im2Fourier(image_slice, mode=mode)
        transformed_slices.append(transformed_slice)

    return np.stack(transformed_slices, axis=0)


def Fourier2Vol(fourier_volume: np.ndarray, mode: str = "complex") -> np.ndarray:
    """
    Applies the 2D Inverse Fourier Transform to each slice of a transformed volume,
    reconstructing the original volume.

    Args:
        fourier_volume (np.ndarray): The input volume with the Fourier transform applied,
                                     with shape (Z, N, Y, X).
        mode (str): The format of the input. Must be one of "complex",
                    "real", or "freq".

    Returns:
        np.ndarray: The reconstructed volume with shape (Z, C, Y, X).
    """
    if fourier_volume.ndim != 4:
        raise ValueError("Input volume must be a 4D array with shape (Z, N, Y, X).")

    Z, _, Y, X = fourier_volume.shape
    reconstructed_slices = []

    for z in range(Z):
        fourier_slice = fourier_volume[z, :, :, :]
        reconstructed_slice = Fourier2Im(fourier_slice, mode=mode)
        reconstructed_slices.append(reconstructed_slice)

    return np.stack(reconstructed_slices, axis=0)


def VolDiscretizer(volume: np.ndarray, n: int = 3) -> np.ndarray:
    """
    Applies discretization to each image (slice) of a volume,
    quantizing float values to n integer ranges.

    Args:
        volume (np.ndarray): The input volume as a 4D array with
                             shape (Z, C, Y, X).
        n (int): The number of integer levels to quantize to
                 (e.g., 3 for 0, 1, 2).

    Returns:
        np.ndarray: The new volume with discretized values as a 4D integer array.
    """
    if volume.ndim != 4:
        raise ValueError("Input volume must be a 4D array with shape (Z, C, Y, X).")

    Z, C, Y, X = volume.shape
    discretized_slices = []

    for z in range(Z):
        image_slice = volume[z, :, :, :]
        discretized_channels = []
        for c in range(C):
            channel_slice = image_slice[c, :, :]
            discretized_channel = discretizer2n(channel_slice, n=n)
            discretized_channels.append(discretized_channel)

        discretized_slices.append(np.stack(discretized_channels, axis=0))

    return np.stack(discretized_slices, axis=0)


def FreqNorm(fourier_image: np.ndarray) -> np.ndarray:
    """
    Normalizes the magnitude and phase from the output of Im2Fourier with mode 'freq'.

    The magnitude is normalized using a logarithmic scale followed by a Min-Max
    normalization to the range [0, 1]. The phase is scaled to the range [0, 1].

    Args:
        fourier_image (np.ndarray): The input Fourier-transformed matrix with shape
                                     (2C, Y, X), where even channels are magnitude
                                     and odd channels are phase.

    Returns:
        np.ndarray: The output matrix with the same shape, but with normalized
                    magnitude and phase channels.

    Raises:
        ValueError: If the number of channels is not even.
    """
    if fourier_image.ndim != 3:
        raise ValueError("Input image must be a 3D array with shape (N, Y, X).")

    N, Y, X = fourier_image.shape
    if N % 2 != 0:
        raise ValueError(f"The number of channels ({N}) must be even for mode 'freq'.")

    C = N // 2
    normalized_channels = []

    for i in range(C):
        # Separate magnitude and phase for each channel
        magnitude = fourier_image[2 * i, :, :]
        phase = fourier_image[2 * i + 1, :, :]

        # Magnitude normalization: logarithm + Min-Max
        log_magnitude = np.log1p(magnitude)
        min_log_mag = np.min(log_magnitude)
        max_log_mag = np.max(log_magnitude)

        if max_log_mag == min_log_mag:
            norm_magnitude = np.zeros_like(log_magnitude)
        else:
            norm_magnitude = (log_magnitude - min_log_mag) / (max_log_mag - min_log_mag)

        # Phase normalization: Scale to the range [0, 1]
        # np.angle returns phase in the range [-pi, pi]
        norm_phase = (phase + np.pi) / (2 * np.pi)

        normalized_channels.append(norm_magnitude)
        normalized_channels.append(norm_phase)

    return np.stack(normalized_channels, axis=0)


def VolFreqNorm(fourier_volume: np.ndarray) -> np.ndarray:
    """
    Applies frequency-based normalization to each slice of a Fourier-transformed volume.

    This function normalizes the magnitude and phase for each 2D image (slice)
    within a 4D volume, using the same logic as FreqNorm.

    Args:
        fourier_volume (np.ndarray): The input volume with the Fourier transform applied,
                                     with shape (Z, 2C, Y, X).

    Returns:
        np.ndarray: The normalized volume with the same shape.

    Raises:
        ValueError: If the input volume is not a 4D array or if the number of
                    channels is not even.
    """
    if fourier_volume.ndim != 4:
        raise ValueError("Input volume must be a 4D array with shape (Z, N, Y, X).")

    Z, N, Y, X = fourier_volume.shape
    if N % 2 != 0:
        raise ValueError(f"The number of channels ({N}) must be even for mode 'freq'.")

    normalized_slices = []

    for z in range(Z):
        fourier_slice = fourier_volume[z, :, :, :]
        normalized_slice = FreqNorm(fourier_slice)
        normalized_slices.append(normalized_slice)

    return np.stack(normalized_slices, axis=0)


def transform_files(
    path_data: str,
    path_out: str,
    function: Callable[[np.ndarray], np.ndarray],
    mode: str = "complex",
    shape: str = "CYX",
) -> None:
    """
    Applies a function to all image files in a directory and saves the modified images.

    This function iterates through all files in the specified input directory,
    loads each image, applies a user-defined function to its data, and then
    saves the modified image to a new output directory.

    Args:
        path_data (str): The path to the directory containing the input image files.
        path_out (str): The path to the directory where the modified images will be saved.
        function (Callable[[np.ndarray], np.ndarray]): A function that takes a NumPy array
                                                        (the image data) and returns a
                                                        NumPy array (the transformed data).
        shape (str, optional): The dimension order of the image to be loaded.
                                Defaults to 'CYX'.

    Raises:
        TypeError: If path_data, path_out, or shape are not strings.
        TypeError: If the 'function' argument is not a callable object.
        ValueError: If a provided path does not exist.
    """

    if (
        not isinstance(path_data, str)
        or not isinstance(path_out, str)
        or not isinstance(shape, str)
    ):
        raise TypeError("path_data, path_out, and shape must be strings.")

    if not isinstance(function, Callable):
        raise TypeError("'function' must be a callable object.")

    if not os.path.isdir(path_data):
        raise ValueError(f"The input directory '{path_data}' does not exist.")

    if not os.path.exists(path_out):
        print(f"Creating output directory: {path_out}")
        os.makedirs(path_out)

    files = os.listdir(path_data)
    if not files:
        print(f"Warning: No files found in the directory: {path_data}")
        return

    for file in tqdm(files):
        try:
            im = BioImage(os.path.join(path_data, file)).get_image_data(shape, T=0)
            c_im = function(im, mode)
            OmeTiffWriter.save(c_im, os.path.join(path_out, file), dim_order=shape)
        except Exception as e:
            print(f"Error processing file '{file}': {e}")



import numpy as np
from skimage.morphology import remove_small_objects, remove_small_holes
from skimage.measure import label, regionprops
from mmv_im2im.utils.utils import topology_preserving_thinning
import torch
from scipy.ndimage import shift, rotate
import random


def perturb_image(
    im_input,
    opts,
    gaussian_std=0.01,
    sp_prob=0.01,
    speckle_std=0.1,
    color_jitter_factor=0.1,
    max_shift=2,
    max_angle=2,
    scale_range=(0.98, 1.02),
    dropout_rate=0.02,
):
    """
    Applies a random combination of small perturbations to an image (C, X, Y) in NumPy.

    This function is intended for Data Augmentation *before* the model inference
    to generate slightly varied inputs for Monte Carlo Dropout or similar methods.

    Args:
        im_input (np.ndarray): Input image with shape (channels, height, width).
                               Assumed to be a float array (e.g., normalized 0.0 to 1.0).
        opts: List with the option tranformations or string indicatting 'all' ranodm text-> all
        gaussian_std (float): Standard deviation for Gaussian Noise.
        sp_prob (float): Probability for Salt and Pepper Noise (controls the density).
        speckle_std (float): Standard deviation for Speckle Noise (multiplicative).
        color_jitter_factor (float): Max factor for color perturbation (brightness/contrast).
        max_shift (int): Maximum displacement in pixels (for both X and Y axes).
        max_angle (float): Maximum rotation angle in degrees (e.g., ±2 degrees).
        scale_range (tuple): Range (min, max) of the scaling factor (e.g., 0.98 to 1.02).
        dropout_rate (float): Probability that an individual pixel will be dropped (set to 0).

    Returns:
        np.ndarray: The perturbed image.
    """

    # Create a copy to avoid modifying the original array
    im_out = im_input.copy()
    C, X, Y = im_out.shape

    if isinstance(opts, str):
        opts = [
            "gauss_noise",
            "impulse_noise",
            "speckle_noise",
            "color_jitter",
            "shift",
            "rotation",
            "pixel_dropout",
        ]

    # Gaussian Noise (Additive)
    def add_gaussian_noise(img):
        # Generate Gaussian noise array
        noise = np.random.normal(0, gaussian_std, img.shape)
        return img + noise

    # Salt and Pepper Noise (Impulse Noise)
    def add_salt_and_pepper_noise(img):
        out = img.copy()
        # Calculate number of salt (max value) and pepper (min value) points
        num_sp_points = int(sp_prob * X * Y * C / 2)  # Divide by 2 for Salt and Pepper

        # Salt Noise: max values (assuming normalized data 0.0 to 1.0)
        coords_salt = [np.random.randint(0, s, num_sp_points) for s in img.shape]
        out[tuple(coords_salt)] = 1.0

        # Pepper Noise: min values
        coords_pepper = [np.random.randint(0, s, num_sp_points) for s in img.shape]
        out[tuple(coords_pepper)] = 0.0

        return out

    # Speckle Noise (Multiplicative Noise)
    def add_speckle_noise(img):
        # Generate multiplicative noise component
        noise = np.random.normal(0, speckle_std, img.shape)
        return img * (1 + noise)

    # Color Jitter (Small random change in brightness/contrast)
    def apply_color_jitter(img):
        # Choose a random small scale factor (contrast)
        scale = 1.0 + np.random.uniform(-color_jitter_factor, color_jitter_factor)
        # Choose a random small offset (brightness)
        offset = np.random.uniform(-color_jitter_factor / 5, color_jitter_factor / 5)

        # Apply transformation: img * scale + offset
        return (img * scale) + offset

    # Shift (Translation)
    def apply_shift(img):
        # Random shift by 1 or 2 pixels in X and Y (axes 1 and 2)
        shift_x = np.random.randint(-max_shift, max_shift + 1)
        shift_y = np.random.randint(-max_shift, max_shift + 1)

        # Shift is applied to axes X and Y (1 and 2). Channel (axis 0) shift is 0.
        return shift(img, (0, shift_x, shift_y), mode="nearest")

    # Rotation
    def apply_rotation(img):
        # Random very small angle (e.g., ±1° or ±2°)
        angle = np.random.uniform(-max_angle, max_angle)

        # Rotate in the X-Y plane (axes 1 and 2).
        # reshape=False ensures the output shape is the same as input.
        return rotate(img, angle, axes=(1, 2), reshape=False, mode="nearest")

    # Pixel Dropout (Setting random pixels to 0)
    def apply_pixel_dropout(img):
        # Create a binary mask where True (1) pixels are kept and False (0) are dropped
        mask = np.random.binomial(1, 1 - dropout_rate, size=img.shape)
        return img * mask

    # List of all transformation functions
    # Each function takes im_out and returns the transformed array
    transformations = []
    if "gauss_noise" in opts:
        transformations.append(add_gaussian_noise)
    if "impulse_noise" in opts:
        transformations.append(add_salt_and_pepper_noise)
    if "speckle_noise" in opts:
        transformations.append(add_speckle_noise)
    if "color_jitter" in opts:
        transformations.append(apply_color_jitter)
    if "shift" in opts:
        transformations.append(apply_shift)
    if "rotation" in opts:
        transformations.append(apply_rotation)
    if "pixel_dropout" in opts:
        transformations.append(apply_pixel_dropout)

    if len(transformations) == 0:
        raise ValueError("Invalid transformations")

    # --- Random Application of Transformations ---

    # Choose how many transformations to apply (1 to N, where N is the total number of defined transforms)
    num_transforms_to_apply = random.randint(1, len(transformations))

    # Randomly select the transformations (without replacement)
    selected_transforms = random.sample(transformations, num_transforms_to_apply)

    # Apply the selected transformations in random order
    for transform_func in selected_transforms:
        im_out = transform_func(im_out)

    # Crucial: Ensure data remains within a valid range (e.g., 0.0 to 1.0)
    # The clip operation prevents extreme values generated by noise/jitter from breaking the model.
    # Note: If your input images are not normalized to [0, 1], adjust this clipping range accordingly.
    im_out = np.clip(im_out, 0.0, 1.0)

    return im_out


def Perycites_correction(seg_full):
    seg_2 = remove_small_objects(seg_full == 2, min_size=30)
    seg_2_mid = np.logical_xor(seg_2, remove_small_objects(seg_2, min_size=300))

    for zz in range(seg_2_mid.shape[0]):
        seg_label, num_obj = label(seg_2_mid[zz, :, :], return_num=True)
        if num_obj > 0:
            stats = regionprops(seg_label)
            for ii in range(num_obj):
                if (
                    stats[ii].eccentricity < 0.88
                    and stats[ii].solidity > 0.85
                    and stats[ii].area < 150
                ):
                    seg_z = seg_2[zz, :, :]
                    seg_z[seg_label == (ii + 1)] = 0
                    seg_2[zz, :, :] = seg_z

    seg_full[seg_full == 2] = 1
    seg_full[seg_2 > 0] = 2
    return seg_full


def Remove_objects(seg_full, n_classes, remove_object_size, voxel_sizes=(1, 1, 1)):
    """
    Applies removal of small objects to all object classes (1 to n_classes-1)
    in a 3D segmentation volume, allowing for class-specific size thresholds.

    Args:
        seg_full (np.ndarray): The 3D segmentation volume with integer class values.
        n_classes (int): The total number of classes in the segmentation (including background).
        remove_object_size (list or int): A single minimum size (int) or a list of
                                          minimum sizes (list) for objects to be kept.
                                          If a list, its length must be 1 or equal
                                          to the number of object classes (n_classes - 1).

    Returns:
        np.ndarray: The segmentation volume with small objects removed for each class.

    Raises:
        ValueError: If the length of remove_object_size list is invalid.
    """
    pz, py, px = voxel_sizes
    voxel_volume = pz * py * px

    classes_to_process = range(1, n_classes)
    num_target_classes = len(classes_to_process)

    thresholds = []

    if not isinstance(remove_object_size, list):
        physical_thresholds = [remove_object_size] * num_target_classes
    else:
        list_len = len(remove_object_size)

        if list_len == 1:
            physical_thresholds = [remove_object_size[0]] * num_target_classes

        elif list_len == num_target_classes:
            physical_thresholds = remove_object_size

        else:
            raise ValueError(
                f"The list 'remove_object_size' has {list_len} elements, "
                f"but {num_target_classes} (or 1) were expected for the {num_target_classes} classes to process "
                f"(Class 1 to {n_classes - 1}). The background (Class 0) is ignored."
            )

    for physical_size in physical_thresholds:
        min_voxel_count = int(np.ceil(physical_size / voxel_volume))
        thresholds.append(min_voxel_count)

    seg_cleaned = np.zeros_like(seg_full)

    for i, class_id in enumerate(classes_to_process):
        min_size_threshold = thresholds[i]

        seg_class_mask = seg_full == class_id

        if seg_class_mask.any():
            seg_class_clean = remove_small_objects(
                seg_class_mask, min_size=min_size_threshold
            )
            seg_cleaned[seg_class_clean] = class_id

    return seg_cleaned


def Hole_Correction(seg_full, n_classes, hole_size_threshold, voxel_sizes=(1, 1, 1)):
    """
    Applies hole correction to multiple classes in a segmentation volume.

    The correction is applied to object classes (typically 1 up to n_classes-1).
    Each class can have a different hole size threshold. It also includes
    an initial removal of small objects for all classes.

    Args:
        seg_full (np.ndarray): The 3D segmentation volume with integer class values.
        n_classes (int): The total number of classes in the segmentation (including background).
        hole_size_threshold (list or int): A single threshold (int) or a list of
                                           thresholds (list) for hole correction.
                                           If a list, its length must be 1 or
                                           equal to the number of object classes (n_classes - 1).

    Returns:
        np.ndarray: The corrected segmentation volume.

    Raises:
        ValueError: If the length of hole_size_threshold is not 1 and is less than n_classes - 1.
    """
    pz, py, px = voxel_sizes
    pixel_area = py * px

    classes_to_correct = range(1, n_classes)
    num_target_classes = len(classes_to_correct)

    thresholds = []

    if not isinstance(hole_size_threshold, list):
        physical_thresholds = [hole_size_threshold] * num_target_classes
    else:
        list_len = len(hole_size_threshold)

        if list_len == 1:
            physical_thresholds = [hole_size_threshold[0]] * num_target_classes

        elif list_len == num_target_classes:
            physical_thresholds = hole_size_threshold

        else:
            raise ValueError(
                f"The list 'hole_size_threshold' has {list_len} elements, "
                f"but {num_target_classes} (or 1) were expected for the {num_target_classes} classes to correct "
                f"(Class 1 to {n_classes - 1}). The background (Class 0) does not need a threshold."
            )

    for physical_area in physical_thresholds:
        area_threshold = int(np.ceil(physical_area / pixel_area))
        thresholds.append(area_threshold)

    seg_corrected = seg_full.copy()

    for i, class_id in enumerate(classes_to_correct):
        threshold = thresholds[i]
        seg_obj_mask = seg_corrected == class_id

        if seg_obj_mask.any():
            seg_obj_slice_corrected = seg_obj_mask.copy()

            for zz in range(seg_full.shape[0]):
                s_v = remove_small_holes(
                    seg_obj_slice_corrected[zz, :, :], area_threshold=threshold
                )
                seg_obj_slice_corrected[zz, :, :] = s_v[:, :]
            seg_corrected[seg_corrected == class_id] = 0
            seg_corrected[seg_obj_slice_corrected] = class_id

    return seg_corrected


def Thickness_Corretion(
    seg_full, n_classes, min_thickness_physical, voxel_sizes=(1, 1, 1)
):
    """
    Applies topology-preserving thinning (thickness correction) to all object
    classes (1 to n_classes-1) in a 3D segmentation volume, using a specific
    minimum thickness for each class. Class 0 (background) is automatically ignored.

    Args:
        seg_full (np.ndarray): The 3D segmentation volume with integer class values.
        n_classes (int): The total number of classes in the segmentation (including Class 0).
        min_thickness_physical (list or np.ndarray): A list or array of minimum
                                                thickness values. The index 'i'
                                                corresponds to the minimum thickness for Class i+1.
                                                (e.g., index 0 is for Class 1).

    Returns:
        np.ndarray: The segmentation volume where each object class has been thinned,
                    preserving its original class label.
    """
    pz, py, px = voxel_sizes
    distance_unit = (py + px) / 2

    classes_to_process = range(1, n_classes)
    num_object_classes = len(classes_to_process)

    # 1. Validate the length of the minimum thickness list
    if len(min_thickness_physical) != num_object_classes:
        raise ValueError(
            f"The length of 'min_thickness_list' ({len(min_thickness_physical)}) does not match "
            f"the number of object classes to process ({num_object_classes}). "
            "Class 0 (background) is ignored, so {num_object_classes} values are expected "
            " (one for each class from 1 to {n_classes-1})."
        )

    min_thickness_list = []
    for physical_distance in min_thickness_physical:
        min_thickness_pixel = int(np.ceil(physical_distance / distance_unit))
        min_thickness_list.append(min_thickness_pixel)

    seg_corrected = np.zeros_like(seg_full)

    for i, class_id in enumerate(classes_to_process):

        current_min_thickness = min_thickness_list[i]
        seg_class_mask = seg_full == class_id

        seg_thinned = topology_preserving_thinning(
            seg_class_mask, min_thickness=current_min_thickness, thin=1
        )

        seg_corrected[seg_thinned > 0] = class_id

    return seg_corrected


def adjust_volume(volume: np.ndarray) -> np.ndarray:
    """
    Applies the square root (sqrt) to a NumPy volume (2D or 3D) based on
    the percentage of values within specific orders of magnitude.

    Args:
        volume (np.ndarray): The input volume with shape (c, y, x) or (y, x).

    Returns:
        np.ndarray: The transformed volume.
    """

    # 1. Condition: 80% or more of the values are of the order e^-6 or smaller
    # Range: [0, 1e-6]

    threshold_e_neg_6 = 1e-6
    # Create a mask for values less than or equal to 1e-6 (including 0)
    mask_e_neg_6_or_less = np.less_equal(volume, threshold_e_neg_6)

    # Count how many elements meet the condition
    count_e_neg_6_or_less = np.count_nonzero(mask_e_neg_6_or_less)

    # Calculate the percentage
    total_elements = volume.size
    percentage_e_neg_6_or_less = (count_e_neg_6_or_less / total_elements) * 100

    # 2. Condition: 80% or more of the values are of the order 0.000n or 0.00n
    # Range: (1e-6, 0.01] (Strictly greater than 1e-6 and less than or equal to 0.01)

    lower_threshold_00n = 1e-6
    upper_threshold_00n = 1e-2  # 0.01

    mask_00n = np.logical_and(
        np.greater(volume, lower_threshold_00n),
        np.less_equal(volume, upper_threshold_00n),
    )

    # Count how many elements meet the condition
    count_00n = np.count_nonzero(mask_00n)

    # Calculate the percentage
    percentage_00n = (count_00n / total_elements) * 100

    # Apply double sqrt (e^{-6} or less)
    if percentage_e_neg_6_or_less >= 80:
        volume = np.sqrt(np.sqrt(volume))

    # Apply single sqrt (0.000n or 0.00n)
    elif percentage_00n >= 80:
        # sqrt(x)
        volume = np.sqrt(volume)

    else:
        return volume

    return volume


def Extract_Uncertainty_Maps(
    logits_samples, compute_mode, relative_MI=True, var_reductor=True, estabilizer=False
):
    """
    Generates an uncertainty map based on the compute_mode.

    Args:
        logits_samples (list[np.ndarray]): List of N logit samples, where each
                                           sample has shape (C, Y, X).
        compute_mode (str): Uncertainty calculation mode:
                            'variance' (Returns variance) or
                            'mutual_inf' (Returns Mutual Information).
                            'entropy' (Returns Total uncertainity)
                            'prob_in' (Returns 1-Prob)
        relative_MI (bool): If True and compute_mode='mutual_inf', MI is normalized
                            by ln(C) to the range [0, 1]. Ignored for 'variance'.
        var_reductor (bool): ONLY APPLIES TO 'variance' mode.
                             If True, returns the minimum variance across classes
                             (shape: (Y, X)).
                             If False, returns the variance for all classes
                             (shape: (C, Y, X)).

    Returns:
        np.ndarray: The resulting uncertainty map. Shape is (Y, X) for 'mutual_inf'
                    and reduced 'variance', or (C, Y, X) for unreduced 'variance'.

    Raises:
        ValueError: If the list of logit samples is empty or the computation mode is invalid.
    """

    if not logits_samples:
        raise ValueError("The list of logit samples cannot be empty.")

    # Convert Logits to Probabilities ---

    # Stack samples along a new axis (axis 0). Shape: (N_samples, C, Y, X)
    stacked_logits = np.stack(logits_samples, axis=0)
    N, C, Y, X = stacked_logits.shape

    # Apply Softmax along the class axis (axis 1) to get probabilities P.
    logits_tensor = torch.from_numpy(stacked_logits).float()

    # stacked_probs.shape: (N_samples, C, Y, X)
    stacked_probs = torch.nn.functional.softmax(logits_tensor, dim=1).numpy()

    # CALCULATE AND MERGE UNCERTAINTY ---

    if compute_mode == "variance":
        # Calculate Variance of probabilities (class-wise uncertainty)
        # uncertainty_map_split shape: (C, Y, X)
        uncertainty_map_split = np.var(stacked_probs, axis=0)

        if var_reductor:
            # Merge: Take the MINIMUM uncertainty across classes (axis 0)
            # merged_uncertainty shape: (Y, X)
            merged_uncertainty = np.min(uncertainty_map_split, axis=0)

            if estabilizer:
                merged_uncertainty = adjust_volume(merged_uncertainty)
            return merged_uncertainty
        else:
            # Return per-class uncertainty map
            # shape: (C, Y, X)
            if estabilizer:
                uncertainty_map_split = adjust_volume(uncertainty_map_split)
            return uncertainty_map_split
    elif compute_mode == "prob_inv":
        # Probability map
        # stacked_max_probs.shape: (N_samples, Y, X)
        prob_comp = 1 - np.max(stacked_probs, axis=1)
        # averaged_max_probs.shape: (Y, X)
        prob_comp = np.mean(prob_comp, axis=0)

        if estabilizer:
            prob_comp = adjust_volume(prob_comp)

        return prob_comp
    elif compute_mode == "mutual_inf" or compute_mode == "entropy":
        # Calculate Mutual Information (MI)

        # helps to avoid log(0)
        epsilon = 1e-12

        # a) Average Predictive Probability (E[P(y|x)])
        avg_probs = np.mean(stacked_probs, axis=0)
        # Cliping para la entropía total (H[E[P(y|x)]])
        avg_probs_clipped = np.clip(avg_probs, a_min=epsilon, a_max=None)

        # b) Total Predictive Entropy (H[E[P(y|x)]])
        entropy_total = -np.sum(avg_probs_clipped * np.log(avg_probs_clipped), axis=0)

        # c) Average Conditional Entropy (E[H[P(y|x, w)]])

        # Cliping para la entropía condicional (P(y|x,w) * log(P(y|x,w)))
        stacked_probs_clipped = np.clip(stacked_probs, a_min=epsilon, a_max=None)

        # Cálculo de la entropía por muestra: -sum(P log P)
        per_sample_entropy = -np.sum(
            stacked_probs_clipped * np.log(stacked_probs_clipped), axis=1
        )

        # Promedio de la entropía por muestra (E[H[P(y|x, w)]])
        entropy_avg_conditional = np.mean(per_sample_entropy, axis=0)

        if compute_mode == "entropy":
            # Entropy
            mutual_information_map = entropy_total
        else:
            # d) Mutual Information (MI)
            mutual_information_map = entropy_total - entropy_avg_conditional

        #  Apply Relative Normalization if requested
        if relative_MI:
            # Normalization factor: ln(C), the max possible entropy (MI max theoretical bound)
            max_mi = np.log(C)

            if max_mi == 0:  # Handle C=1 case (though unlikely for segmentation)
                return np.zeros_like(mutual_information_map)

            # Clip result to ensure strict [0, 1] range due to floating point arithmetic
            normalized_mi = np.clip(mutual_information_map / max_mi, 0.0, 1.0)

            if estabilizer:
                normalized_mi = adjust_volume(normalized_mi)

            return normalized_mi
        else:
            # Return original MI (in nats), range [0, ln(C)]
            if estabilizer:
                mutual_information_map = adjust_volume(mutual_information_map)
            return mutual_information_map

    else:
        raise ValueError("Invalid computation mode.")



# adapted from https://github.com/juglab/EmbedSeg/tree/main/EmbedSeg/criterions
import torch
import torch.nn as nn
from mmv_im2im.utils.lovasz_losses import lovasz_hinge


class SpatialEmbLoss_3d(nn.Module):
    def __init__(
        self,
        grid_z=32,
        grid_y=1024,
        grid_x=1024,
        pixel_z=1,
        pixel_y=1,
        pixel_x=1,
        n_sigma=3,
        foreground_weight=10,
        use_costmap=False,
    ):
        super().__init__()

        print(
            f"Created spatial emb loss function with: n_sigma: {n_sigma},"
            f"foreground_weight: {foreground_weight}"
        )
        print("*************************")
        self.n_sigma = n_sigma
        self.foreground_weight = foreground_weight

        xm = (
            torch.linspace(0, pixel_x, grid_x)
            .view(1, 1, 1, -1)
            .expand(1, grid_z, grid_y, grid_x)
        )
        ym = (
            torch.linspace(0, pixel_y, grid_y)
            .view(1, 1, -1, 1)
            .expand(1, grid_z, grid_y, grid_x)
        )
        zm = (
            torch.linspace(0, pixel_z, grid_z)
            .view(1, -1, 1, 1)
            .expand(1, grid_z, grid_y, grid_x)
        )
        xyzm = torch.cat((xm, ym, zm), 0)

        self.register_buffer("xyzm", xyzm)

        # TODO: currently, the costmap for embedding loss needs
        # further investigation, so set to False until fixed
        self.use_costmap = False  # use_costmap

    def forward(
        self,
        prediction,
        instances,
        labels,
        center_images,
        costmaps=None,
        w_inst=1,
        w_var=10,
        w_seed=1,
    ):
        # instances B 1 Z Y X
        batch_size, depth, height, width = (
            prediction.size(0),
            prediction.size(2),
            prediction.size(3),
            prediction.size(4),
        )

        # weighted loss
        # instances_adjusted = instances
        instances_adjusted = instances.clone().detach()
        if self.use_costmap:
            # make sure costmaps only work as exclusion masks, i.e.,
            # only containing values of 0 and 1, 0 = pixels to exclude
            costmaps[costmaps > 1] = 1
            costmaps[costmaps > 0] = 1
            assert not torch.any(costmaps < 0), "cannot have negative value in costmap"

            # only need to adjust instances in this step, because for pixels
            # with zero weight, this step will ignore the corresponding instances
            instances_adjusted[costmaps == 0] = 0

        xyzm_s = self.xyzm[:, 0:depth, 0:height, 0:width].contiguous()  # 3 x d x h x w

        loss = 0

        for b in range(0, batch_size):
            spatial_emb = torch.tanh(prediction[b, 0:3]) + xyzm_s  # 3 x d x h x w
            sigma = prediction[b, 3 : 3 + self.n_sigma]  # n_sigma x d x h x w
            seed_map = torch.sigmoid(
                prediction[b, 3 + self.n_sigma : 3 + self.n_sigma + 1]
            )  # 1 x d x h x w
            # loss accumulators
            var_loss = 0
            instance_loss = 0
            seed_loss = 0
            obj_count = 0

            if self.use_costmap:
                costmap = costmaps[b]
            instance = instances[b]  # without costmap adjustment
            label = labels[b]
            center_image = center_images[b]

            # use adjusted instance to find all ids
            # i.e., if one instance is completed inside the
            # exclusion mask, then this instance will be skipped
            instance_ids = instances_adjusted[b].unique()
            instance_ids = instance_ids[instance_ids != 0]

            # regress bg to zero
            bg_mask = label == 0

            # adjust the cost here, because some of the background pixels might
            # have zero weight. All we need to do is to make sure the pixels
            # being exluded not contributing to the contribution, i.e., with
            # value False in bg_mask
            if self.use_costmap:
                bg_mask[costmap == 0] = 0
            if bg_mask.sum() > 0:
                seed_loss += torch.sum(torch.pow(seed_map[bg_mask] - 0, 2))

            for id in instance_ids:
                in_mask = instance.eq(id)  # 1 x d x h x w

                center_mask = in_mask & center_image
                if center_mask.sum().eq(1):
                    center = xyzm_s[center_mask.expand_as(xyzm_s)].view(3, 1, 1, 1)
                else:
                    xyz_in = xyzm_s[in_mask.expand_as(xyzm_s)].view(3, -1)
                    center = xyz_in.mean(1).view(3, 1, 1, 1)  # 3 x 1 x 1 x 1

                # calculate sigma
                sigma_in = sigma[in_mask.expand_as(sigma)].view(
                    self.n_sigma, -1
                )  # 3 x N

                smean = sigma_in.mean(1).view(
                    self.n_sigma, 1, 1, 1
                )  # n_sigma x 1 x 1 x 1

                # calculate var loss before exp
                var_loss = var_loss + torch.mean(
                    torch.pow(sigma_in - smean[..., 0, 0].detach(), 2)
                )

                s = torch.exp(smean * 10)
                dist = torch.exp(
                    -1
                    * torch.sum(torch.pow(spatial_emb - center, 2) * s, 0, keepdim=True)
                )

                # apply lovasz-hinge loss
                instance_loss = instance_loss + lovasz_hinge(dist * 2 - 1, in_mask)

                # seed loss
                seed_loss += self.foreground_weight * torch.sum(
                    torch.pow(seed_map[in_mask] - dist[in_mask].detach(), 2)
                )

                obj_count += 1

            if obj_count > 0:
                instance_loss /= obj_count
                var_loss /= obj_count

            # seed_loss = seed_loss / (depth * height * width)
            if self.use_costmap:
                seed_loss = seed_loss / costmap.sum()
            else:
                seed_loss = seed_loss / (depth * height * width)

            loss += w_inst * instance_loss + w_var * var_loss + w_seed * seed_loss

        loss = loss / (b + 1)

        return loss + prediction.sum() * 0


class SpatialEmbLoss_2d(nn.Module):
    def __init__(
        self,
        grid_y=1024,
        grid_x=1024,
        pixel_y=1,
        pixel_x=1,
        n_sigma=2,
        foreground_weight=10,
        use_costmap=False,
    ):
        super().__init__()

        print(
            f"Created spatial emb loss function with: n_sigma: {n_sigma},"
            f"foreground_weight: {foreground_weight}"
        )
        print("*************************")
        self.n_sigma = n_sigma
        self.foreground_weight = foreground_weight

        # coordinate map
        xm = torch.linspace(0, pixel_x, grid_x).view(1, 1, -1).expand(1, grid_y, grid_x)
        ym = torch.linspace(0, pixel_y, grid_y).view(1, -1, 1).expand(1, grid_y, grid_x)
        xym = torch.cat((xm, ym), 0)
        self.register_buffer("xym", xym)

        # TODO: currently, the costmap for embedding loss needs
        # further investigation, so set to False until fixed
        self.use_costmap = False  # use_costmap

    def forward(
        self,
        prediction,
        instances,
        labels,
        center_images,
        costmaps=None,
        w_inst=1,
        w_var=10,
        w_seed=1,
    ):
        # instances B C Y X
        batch_size, height, width = (
            prediction.size(0),
            prediction.size(2),
            prediction.size(3),
        )
        xym_s = self.xym[:, 0:height, 0:width].contiguous()  # 2 x h x w

        # weighted loss
        instances_adjusted = instances.clone().detach()
        if self.use_costmap:
            # make sure costmaps only work as exclusion masks, i.e.,
            # only containing values of 0 and 1, 0 = pixels to exclude
            costmaps[costmaps > 1] = 1
            costmaps[costmaps > 0] = 1
            assert not torch.any(costmaps < 0), "cannot have negative value in costmap"

            # only need to adjust instances in this step, because for pixels with
            # zero weight, this step will ignore the corresponding instances
            instances_adjusted[costmaps == 0] = 0

        loss = 0

        for b in range(0, batch_size):
            spatial_emb = torch.tanh(prediction[b, 0:2]) + xym_s  # 2 x h x w #TODO
            sigma = prediction[b, 2 : 2 + self.n_sigma]  # n_sigma x h x w
            seed_map = torch.sigmoid(
                prediction[b, 2 + self.n_sigma : 2 + self.n_sigma + 1]
            )  # 1 x h x w

            # loss accumulators
            var_loss = 0
            instance_loss = 0
            seed_loss = 0
            obj_count = 0

            if self.use_costmap:
                costmap = costmaps[b]
            instance = instances[b]  # without adjustment
            label = labels[b]
            center_image = center_images[b] > 0

            # use adjusted instance to find all ids
            instance_ids = instances_adjusted[b].unique()
            instance_ids = instance_ids[instance_ids != 0]

            # regress bg to zero
            bg_mask = label == 0

            # adjust the cost here, because some of the background pixels might
            # have zero weight. All we need to do is to make sure the pixels
            # being exluded not contributing to the contribution, i.e., with
            # value False in bg_mask
            if self.use_costmap:
                bg_mask[costmap == 0] = 0

            if bg_mask.sum() > 0:
                seed_loss += torch.sum(torch.pow(seed_map[bg_mask] - 0, 2))

            for id in instance_ids:
                in_mask = instance.eq(id)  # 1 x h x w

                center_mask = in_mask & center_image
                if center_mask.sum().eq(1):
                    center = xym_s[center_mask.expand_as(xym_s)].view(2, 1, 1)
                else:
                    xy_in = xym_s[in_mask.expand_as(xym_s)].view(
                        2, -1
                    )  # TODO --> should this edge case change!
                    center = xy_in.mean(1).view(2, 1, 1)  # 2 x 1 x 1

                # calculate sigma
                sigma_in = sigma[in_mask.expand_as(sigma)].view(self.n_sigma, -1)

                s = sigma_in.mean(1).view(self.n_sigma, 1, 1)  # n_sigma x 1 x 1

                # calculate var loss before exp
                var_loss = var_loss + torch.mean(
                    torch.pow(sigma_in - s[..., 0].detach(), 2)
                )

                s = torch.exp(s * 10)  # TODO
                dist = torch.exp(
                    -1
                    * torch.sum(torch.pow(spatial_emb - center, 2) * s, 0, keepdim=True)
                )

                # apply lovasz-hinge loss
                instance_loss = instance_loss + lovasz_hinge(dist * 2 - 1, in_mask)

                # seed loss
                seed_loss += self.foreground_weight * torch.sum(
                    torch.pow(seed_map[in_mask] - dist[in_mask].detach(), 2)
                )

                obj_count += 1

            if obj_count > 0:
                instance_loss /= obj_count
                var_loss /= obj_count

            # seed_loss = seed_loss / (height * width)
            if self.use_costmap:
                seed_loss = seed_loss / costmap.sum()
            else:
                seed_loss = seed_loss / (height * width)

            loss += w_inst * instance_loss + w_var * var_loss + w_seed * seed_loss

        # The sum of the losses of all the instances in the batch.
        loss = loss / (b + 1)

        return loss + prediction.sum() * 0


def calculate_iou(pred, label):
    intersection = ((label == 1) & (pred == 1)).sum()
    union = ((label == 1) | (pred == 1)).sum()
    if not union:
        return 0
    else:
        iou = intersection.item() / union.item()
        return iou



from typing import Union, Dict, List
from pathlib import Path
from functools import partial
import importlib
import numpy as np
import inspect
from bioio import BioImage
from typing import Sequence, Tuple
from monai.data import ImageReader
from monai.utils import ensure_tuple, require_pkg
from monai.config import PathLike
from monai.data.image_reader import _stack_images
import bioio_tifffile


@require_pkg(pkg_name="bioio")
class monai_bio_reader(ImageReader):
    def __init__(self, **kwargs):
        super().__init__()
        self.kwargs = kwargs

    def read(self, data: Union[Sequence[PathLike], PathLike]):
        filenames: Sequence[PathLike] = ensure_tuple(data)
        img_ = []
        for name in filenames:
            try:
                img_.append(BioImage(f"{name}", reader=bioio_tifffile.Reader))

            except Exception:
                try:
                    img_.append(BioImage(f"{name}"))
                except Exception as e:
                    print(f"Error: {e}")
                    print(f"Image {name} failed at read process check the format.")

        return img_ if len(filenames) > 1 else img_[0]

    def get_data(self, img) -> Tuple[np.ndarray, Dict]:
        img_array: List[np.ndarray] = []

        for img_obj in ensure_tuple(img):
            data = img_obj.get_image_data(**self.kwargs)
            img_array.append(data)

        return _stack_images(img_array, {}), {}

    def verify_suffix(self, filename: Union[Sequence[PathLike], PathLike]) -> bool:
        return True


"""
def aicsimageio_reader(fn, **kwargs):
    img = AICSImage(fn).reader.get_image_dask_data(**kwargs)
    img_data = tio.data.io.check_uint_to_int(img.compute())
    return img_data, np.eye(4)


def load_yaml_cfg(yaml_path):
    with open(yaml_path, "r") as stream:
        opt_dict = yaml.safe_load(stream)

    # convert dictionary to attribute-like object
    opt = Munch(opt_dict)

    return opt


def get_max_shape(subjects):
    dataset = tio.SubjectsDataset(subjects)
    shapes = np.array([s.spatial_shape for s in dataset])
    return shapes.max(axis=0)
"""


def parse_config_func_without_params(info):
    # TODO add docstring
    my_module = importlib.import_module(info["module_name"])
    my_func = getattr(my_module, info["func_name"])
    return my_func


def parse_config(info):
    # univeral configuration parser
    my_func = parse_config_func_without_params(info)
    if "params" in info:
        if inspect.isclass(my_func):
            return my_func(**info["params"])
        else:
            return partial(my_func, **info["params"])
    else:
        if inspect.isclass(my_func):
            return my_func()
        else:
            return my_func


def parse_config_func(info):
    # TODO: remove this function, use the universal parse_config above
    my_func = parse_config_func_without_params(info)
    if "params" in info:
        return partial(my_func, **info["params"])
    else:
        return my_func


def parse_ops_list(trans_func: List[Dict]):
    op_list = []
    # {"module_name":"numpy.random",
    #  "func_name":"randint",
    #  "params":{"low":0,"high":1}}
    # load functions according to config
    for trans_dict in trans_func:
        op_list.append(parse_config(trans_dict))
        # trans_module = importlib.import_module(trans_dict["module_name"])
        # trans_func = getattr(trans_module, trans_dict["func_name"])
        # op_list.append(trans_func(**trans_dict["params"]))
    return op_list


def generate_test_dataset_dict(data: Union[str, Path], data_type: str = None) -> List:
    """
    different options for "data":
    - one CSV
    - one folder
    Return
        a list of filename
    """
    dataset_list = []
    data = Path(data).expanduser()
    if data.is_file():
        # should be a csv of dataframe
        import pandas as pd

        df = pd.read_csv(data)
        for row in df.iterrows():
            dataset_list.append(row[data_type])

    elif data.is_dir():
        if "*" in data_type:
            all_filename = sorted(data.glob(data_type))
        else:
            all_filename = sorted(data.glob(f"*{data_type}"))
        assert len(all_filename) > 0, f"no file found in {data}"
        print(f"{len(all_filename)} files are found at {data}")
        all_filename.sort()
        for fn in all_filename:
            dataset_list.append(fn)
    else:
        print(f"{data} is not a valid file or directory")

    return dataset_list


def generate_dataset_dict(data: Union[str, Path, Dict]) -> List[Dict]:
    """
    different options for "data":
    - one CSV (columns: source, target, cmap), then split
    - one folder (_IM.tiff, _GT.tiff, _CM.tiff), then split
    - a dictionary of two or three folders (Im, GT, CM), then split

    Return
        a list of dict, each dict contains 2 or 3 keys
        "source_fn", "target_fn", "costmap_fn" (optional)
    """
    dataset_list = []
    if isinstance(data, str) or isinstance(data, Path):
        data = Path(data).expanduser()
        if data.is_file():
            # should be a csv of dataframe
            import pandas as pd

            df = pd.read_csv(data)
            assert "source_path" in df.columns, "column source_path not found"
            assert "target_path" in df.columns, "column target_path not found"

            # check if costmap is needed
            if "costmap_path" in df.columns:
                cm_flag = True
            else:
                cm_flag = False

            for row in df.itertuples():
                if cm_flag:
                    dataset_list.append(
                        {
                            "source_fn": row.source_path,
                            "target_fn": row.target_path,
                            "costmap_fn": row.costmap_path,
                        }
                    )
                else:
                    dataset_list.append(
                        {
                            "source_fn": row.source_path,
                            "target_fn": row.target_path,
                        }  # noqa E501
                    )
        elif data.is_dir():
            all_filename = sorted(data.glob("*_IM.*"))
            assert len(all_filename) > 0, f"no file found in {data}"

            all_filename.sort()
            for fn in all_filename:
                target_fn = data / fn.name.replace("_IM.", "_GT.")
                costmap_fn = data / fn.name.replace("_IM.", "_CM.")
                if costmap_fn.is_file():
                    dataset_list.append(
                        {
                            "source_fn": fn,
                            "target_fn": target_fn,
                            "costmap_fn": costmap_fn,
                        }
                    )
                else:
                    dataset_list.append(
                        {
                            "source_fn": fn,
                            "target_fn": target_fn,
                        }
                    )
        else:
            print(f"{data} is not a valid file or directory")

    elif isinstance(data, Dict):
        # assume 3~4 keys: "source_dir", "target_dir", and
        # "image_type", "costmap_dir" (optional)
        if "costmap_dir" in data:
            cm_path = Path(data["costmap_dir"]).expanduser()
        else:
            cm_path = None

        source_path = Path(data["source_dir"]).expanduser()
        target_path = Path(data["target_dir"]).expanduser()

        data_type = data["image_type"]

        all_filename = sorted(source_path.glob(f"*.{data_type}"))
        assert len(all_filename) > 0, f"no file found in {source_path}"
        all_filename.sort()

        for fn in all_filename:
            target_fn = target_path / fn.name
            if cm_path is not None:
                costmap_fn = cm_path / fn.name
                dataset_list.append(
                    {
                        "source_fn": fn,
                        "target_fn": target_fn,
                        "costmap_fn": costmap_fn,
                    }  # noqa E501
                )
            else:
                dataset_list.append({"source_fn": fn, "target_fn": target_fn})

    else:
        print("unsupported data type")

    assert len(dataset_list) > 0, f"empty dataset in {data}"

    return dataset_list


def load_one_folder(data) -> List:
    dataset_list = []

    all_filename = sorted(data.glob("*_IM.*"))
    assert len(all_filename) > 0, f"no file found in {data}"

    # parse how many images associated with one subject
    basename = all_filename[0].stem
    basename = basename[: basename.rfind("_")]
    subject_files = sorted(data.glob(f"{basename}_*.*"))
    all_tags = [sfile.stem[sfile.stem.rfind("_") + 1 :] for sfile in subject_files]

    all_filename.sort()
    for fn in all_filename:
        path_list = {}
        for tag_name in all_tags:
            fn_full = data / fn.name.replace("_IM.", f"_{tag_name}.")
            path_list[tag_name] = fn_full
        dataset_list.append(path_list)

    return dataset_list


def load_subfolders(data) -> List:
    dataset_list = []

    # parse how many images associated with one subject
    all_tags = [
        d.name
        for d in sorted(data.iterdir())
        if d.is_dir() and not d.name.startswith(".")
    ]

    data_p1 = data / Path(all_tags[0])
    all_filename = sorted(data_p1.glob("*.*"))
    assert len(all_filename) > 0, f"no file found in {data}"

    all_filename.sort()
    for fn in all_filename:
        path_list = {}
        for tag_name in all_tags:
            fn_full = data / Path(tag_name) / fn.name
            path_list[tag_name] = fn_full
        dataset_list.append(path_list)

    return dataset_list


def generate_dataset_dict_monai(data: Union[str, Path, Dict]) -> List[Dict]:
    """
    different options for "data":
    - one CSV (columns: source, target, cmap), then split
    - one folder (_IM.tiff, _GT.tiff, _CM.tiff), then split
    - one folder with two or three subfolders (Im, GT, CM), then split
    - a dictionary of train/val

    Return
        a list of dict, each dict contains 2 or more keys, such as "IM",
        "GT", and more (optional, e.g. "LM", "CM", etc.)
    """
    if isinstance(data, str):
        try:
            data = eval(data)
        except Exception as e:
            print(f"data path is recognized as a string ... due to {e}")
    dataset_list = []
    if isinstance(data, str) or isinstance(data, Path):
        data = Path(data).expanduser()
        if data.is_file():
            # should be a csv of dataframe

            # TODO: add loading
            pass
        elif data.is_dir():
            if len(sorted(data.glob("*_IM.*"))) > 0:
                dataset_list = load_one_folder(data)
            else:
                # one folder with multiple subfolders (Im, GT, CM)
                dataset_list = load_subfolders(data)
        else:
            print(f"{data} is not a valid file or directory")

    elif isinstance(data, Dict):
        # a dictionary of train/val
        if "train" in data and "val" in data:
            train_path = Path(data["train"])
            train_list = load_one_folder(train_path)
            val_path = Path(data["val"])
            val_list = load_one_folder(val_path)
            dataset_list = {"train": train_list, "val": val_list}
        else:
            raise NotImplementedError(
                "dictionary based loading only takes keywords: train and val"
            )
    else:
        print("unsupported data type")

    assert len(dataset_list) > 0, "empty dataset"

    return dataset_list



# This file was adapted from https://github.com/AllenCellModeling/pytorch_fnet/

from typing import Optional, Tuple, List, Union
import numpy as np
from torch.nn import functional as F
import torch


def norm_around_center(img, z_center: Optional[int] = None, min_z: Optional[int] = 32):
    """Returns normalized version of input img.
    img will be normalized with respect to the mean, std pixel intensity
    of the sub-array of length 32 in the z-dimension centered around the
    img's "z_center".
    Parameters
    ----------
    img
        Input 4D torch.Tensor or numpy array to be normalized.
    z_center
        Z-index of cell centers.
    Returns
    -------
    4D numpy array or tensor
       Normalized image
    """

    # #TODO: Currently, I am not 100 percent sure about how F.pad deals with
    # tensor and ndarray. From the documentation, it seems like only tensors
    # are supported. But, it also works on some kind of ndarray (not all),
    # very strange. Need to follow up to have a better understanding

    if not isinstance(img, np.ndarray):
        img_array = img.numpy()
    else:
        img_array = img.copy()
    img_array = np.squeeze(img_array, axis=0).astype(np.float32)
    if img_array.shape[0] < min_z:
        raise ValueError("Input array must be at least length 32 in first dimension")
    if z_center is None:
        z_center = img_array.shape[0] // 2
    chunk_zlen = 32
    z_start = z_center - chunk_zlen // 2
    if z_start < 0:
        z_start = 0
        print(f"Warning: z_start set to {z_start}")
    if (z_start + chunk_zlen) > img_array.shape[0]:
        z_start = img_array.shape[0] - chunk_zlen
        print(f"Warning: z_start set to {z_start}")
    chunk = img_array[z_start : z_start + chunk_zlen, :, :]
    img = img - chunk.mean()
    img = img / chunk.std()

    return img


def pad_to_multiple(
    img,
    multiple_base: Union[int, Tuple[int], List[int]] = 8,
    pad_value: Union[str, float, int] = 0,
):
    """Returns padded image.
    img will be padded to the size of multiple of "multiple_base".

    Parameters
    ----------
    img:
        Input 4D torch.Tensor or numpy array to be padded, C x X x Y x Z
    multiple_base:
        int or a sequence of int

    Returns
    -------
    4D numpy array or tensor
        padded image
    """

    # #TODO: Currently, I am not 100 percent sure about how F.pad deals with
    # tensor and ndarray. From the documentation, it seems like only tensors
    # are supported. But, it also works on some kind of ndarray (not all),
    # very strange. Need to follow up to have a better understanding

    if not isinstance(img, np.ndarray):
        img_array = img.numpy()
    else:
        img_array = img.copy()

    if isinstance(multiple_base, int):
        if img_array.shape[-1] == 1:
            # Z dim = 1, only pad XY
            pad_base = [multiple_base, multiple_base]
        else:
            # 3D data, pad XYZ
            pad_base = [multiple_base, multiple_base, multiple_base]
    else:
        if img_array.shape[-1] == 1:
            # Z dim = 1, only pad XY
            assert len(multiple_base) == 2, "multiple_base and image shape do not match"
            pad_base = [multiple_base[0], multiple_base[1]]
        else:
            # 3D data, pad XYZ
            assert len(multiple_base) == 3, "multiple_base and image shape do not match"
            pad_base = [multiple_base[0], multiple_base[1], multiple_base[2]]

    multiple_x = img_array.shape[1] // pad_base[0]
    if img_array.shape[1] % pad_base[0] != 0:
        diff_x = pad_base[0] * (multiple_x + 1) - img_array.shape[1]
    else:
        diff_x = 0

    multiple_y = img_array.shape[2] // pad_base[1]
    if img_array.shape[2] % pad_base[1] != 0:
        diff_y = pad_base[1] * (multiple_y + 1) - img_array.shape[2]
    else:
        diff_y = 0

    if len(pad_base) == 3:
        multiple_z = img_array.shape[3] // pad_base[2]
        if img_array.shape[3] % pad_base[2] != 0:
            diff_z = pad_base[2] * (multiple_z + 1) - img_array.shape[3]
        else:
            diff_z = 0
        pad_shape = (
            diff_z // 2,
            diff_z - diff_z // 2,
            diff_y // 2,
            diff_y - diff_y // 2,
            diff_x // 2,
            diff_x - diff_x // 2,
        )
    else:
        pad_shape = (
            0,
            0,
            diff_y // 2,
            diff_y - diff_y // 2,
            diff_x // 2,
            diff_x - diff_x // 2,
        )

    if isinstance(pad_value, str) and pad_value == "mean":
        avg_bg = img_array.mean()
        return F.pad(img, pad_shape, "constant", avg_bg)
    else:
        return F.pad(img, pad_shape, pad_value)


def pad_z(img, target_size: int = 64, pad_value: Union[str, float, int] = 0):
    """Returns padded image.
    img will be padded along z to the size of "target_size".

    Parameters
    ----------
    img:
        Input 4D torch.Tensor or numpy array to be padded, C x Z x Y x X
    target_size:
        int

    Returns
    -------
    4D numpy array or tensor
       padded image
    """

    # #TODO: Currently, I am not 100 percent sure about how F.pad deals with
    # tensor and ndarray. From the documentation, it seems like only tensors
    # are supported. But, it also works on some kind of ndarray (not all),
    # very strange. Need to follow up to have a better understanding

    if not isinstance(img, np.ndarray):
        img_array = img.numpy()
    else:
        img_array = img.copy()

    if img_array.shape[1] < target_size:
        diff_z = target_size - img_array.shape[1]
        pad_shape = (
            0,
            0,
            0,
            0,
            diff_z // 2,
            diff_z - diff_z // 2,
        )
        if isinstance(pad_value, str):
            if pad_value == "mean":
                avg_bg = img_array.mean()
                return F.pad(img, pad_shape, "constant", avg_bg)
            else:
                return F.pad(img, pad_shape, pad_value)
        else:
            print(type(img))
            return F.pad(img, pad_shape, "constant", pad_value)
    else:
        return img


def normalize_staining(
    img, Io=240, alpha=1, beta=0.15, return_unmix_results: bool = False
):
    """Normalize staining appearence of H&E stained images

    Input:
        I: RGB input image (we assume channel order CYX), numpy array or torch.tensor
        Io: (optional) transmitted light intensity
        alpha: parameter from paper (see Reference)
        beta: parameter from paper (see Reference)
        return_unmix_results: whether to return unmixed H and E channel for debugging

    Output:
        Inorm: normalized image
        H: (optional, only when return_unmix_results=True) hematoxylin image
        E: (optional, only when return_unmix_results=True) eosin image

    Reference:
        A method for normalizing histology slides for quantitative analysis. M.
        Macenko et al., ISBI 2009
    """

    HERef = np.array([[0.5626, 0.2159], [0.7201, 0.8012], [0.4062, 0.5581]])

    maxCRef = np.array([1.9705, 1.0308])

    back_to_tensor = False
    original_data_type = None
    if torch.is_tensor(img):
        back_to_tensor = True
        img = img.cpu().numpy()
        original_data_type = img.dtype

    # define height and width of image and move the color dimenstion to the last
    c, h, w = img.shape
    img = np.moveaxis(img, 0, -1)

    assert c == 3, "the first dimension of the image must be 3 (RGB)"

    # reshape image
    img = img.reshape((-1, 3))

    # calculate optical density
    OD = -np.log((img.astype(float) + 1) / Io)

    # remove transparent pixels
    ODhat = OD[~np.any(OD < beta, axis=1)]

    # compute eigenvectors
    eigvals, eigvecs = np.linalg.eigh(np.cov(ODhat.T))

    # eigvecs *= -1

    # project on the plane spanned by the eigenvectors corresponding to the two
    # largest eigenvalues
    That = ODhat.dot(eigvecs[:, 1:3])

    phi = np.arctan2(That[:, 1], That[:, 0])

    minPhi = np.percentile(phi, alpha)
    maxPhi = np.percentile(phi, 100 - alpha)

    vMin = eigvecs[:, 1:3].dot(np.array([(np.cos(minPhi), np.sin(minPhi))]).T)
    vMax = eigvecs[:, 1:3].dot(np.array([(np.cos(maxPhi), np.sin(maxPhi))]).T)

    # a heuristic to make the vector corresponding to hematoxylin first and the
    # one corresponding to eosin second
    if vMin[0] > vMax[0]:
        HE = np.array((vMin[:, 0], vMax[:, 0])).T
    else:
        HE = np.array((vMax[:, 0], vMin[:, 0])).T

    # rows correspond to channels (RGB), columns to OD values
    Y = np.reshape(OD, (-1, 3)).T

    # determine concentrations of the individual stains
    C = np.linalg.lstsq(HE, Y, rcond=None)[0]

    # normalize stain concentrations
    maxC = np.array([np.percentile(C[0, :], 99), np.percentile(C[1, :], 99)])
    tmp = np.divide(maxC, maxCRef)
    C2 = np.divide(C, tmp[:, np.newaxis])

    # recreate the image using reference mixing matrix
    Inorm = np.multiply(Io, np.exp(-HERef.dot(C2)))
    Inorm[Inorm > 255] = 254
    Inorm = np.reshape(Inorm.T, (h, w, 3)).astype(np.uint8)

    # unmix hematoxylin and eosin
    H = np.multiply(
        Io,
        np.exp(
            np.expand_dims(-HERef[:, 0], axis=1).dot(np.expand_dims(C2[0, :], axis=0))
        ),
    )
    H[H > 255] = 254
    H = np.reshape(H.T, (h, w, 3)).astype(np.uint8)

    E = np.multiply(
        Io,
        np.exp(
            np.expand_dims(-HERef[:, 1], axis=1).dot(np.expand_dims(C2[1, :], axis=0))
        ),
    )
    E[E > 255] = 254
    E = np.reshape(E.T, (h, w, 3)).astype(np.uint8)

    # move the color dimension to the first to be consistent with others transforms
    Inorm = np.moveaxis(Inorm, -1, 0)

    if back_to_tensor:
        Inorm = torch.tensor(Inorm.astype(original_data_type))

    if return_unmix_results:
        return Inorm, H, E
    else:
        return Inorm



"""
This module provides lighting module for cycleGAN
"""

import torch
import lightning as pl
from pathlib import Path
import numpy as np
from mmv_im2im.utils.misc import parse_config, parse_config_func
from mmv_im2im.utils.gan_utils import ReplayBuffer
from mmv_im2im.models.nets.gans import define_generator, define_discriminator
from mmv_im2im.utils.model_utils import init_weights
from typing import Dict
from collections import OrderedDict
import itertools
from tifffile import imsave
from skimage.io import imsave as save_rgb


class Model(pl.LightningModule):
    def __init__(self, model_info_xx: Dict, train: bool = True, verbose: bool = False):
        super(Model, self).__init__()

        # Important: This property activates manual optimization.
        self.automatic_optimization = False

        self.verbose = verbose
        self.generator_model_image_AtoB = define_generator(
            model_info_xx.net["generator"]
        )
        self.generator_model_image_BtoA = define_generator(
            model_info_xx.net["generator"]
        )
        self.discriminator_model_image_A = define_discriminator(
            model_info_xx.net["discriminator"]
        )
        self.discriminator_model_image_B = define_discriminator(
            model_info_xx.net["discriminator"]
        )
        if train:
            # init weight
            init_weights(self.generator_model_image_AtoB, init_type="kaiming")
            init_weights(self.generator_model_image_BtoA, init_type="kaiming")
            init_weights(self.discriminator_model_image_A, init_type="kaiming")
            init_weights(self.discriminator_model_image_B, init_type="kaiming")

            # buffer for fake images
            if "fake_pool_size" in model_info_xx.criterion:
                max_fake_pool = model_info_xx.criterion["fake_pool_size"]
            else:
                max_fake_pool = 50
            self.fake_pool_A = ReplayBuffer(max_fake_pool)
            self.fake_pool_B = ReplayBuffer(max_fake_pool)

            # get loss functions
            self.gan_loss = parse_config(model_info_xx.criterion["gan_loss"])
            self.cycle_loss = parse_config(model_info_xx.criterion["cycle_loss"])
            self.identity_loss = parse_config(model_info_xx.criterion["identity_loss"])

            # get weights of different loss
            self.gan_w = model_info_xx.criterion["weights"]["gan_loss"]
            self.cycle_w = model_info_xx.criterion["weights"]["cycle_loss"]
            self.identity_w = model_info_xx.criterion["weights"]["identity_loss"]

            # get info of optimizer and scheduler
            self.optimizer_info = model_info_xx.optimizer
            self.scheduler_info = model_info_xx.scheduler
        else:
            self.direction = model_info_xx.model_extra["inference_direction"]

    def forward(self, x):
        if self.direction == "AtoB":
            return self.generator_model_image_AtoB(x)
        elif self.direction == "BtoA":
            return self.generator_model_image_BtoA(x)

    def configure_optimizers(self):
        generator_optimizer_func = parse_config_func(self.optimizer_info["generator"])
        generator_scheduler_func = parse_config_func(self.scheduler_info["generator"])
        generator_optimizer = generator_optimizer_func(
            itertools.chain(
                self.generator_model_image_AtoB.parameters(),
                self.generator_model_image_BtoA.parameters(),
            )
        )
        generator_scheduler = generator_scheduler_func(generator_optimizer)

        discriminator_optimizer_func = parse_config_func(
            self.optimizer_info["discriminator"]
        )
        discriminator_scheduler_func = parse_config_func(
            self.scheduler_info["discriminator"]
        )
        discriminator_optimizer = discriminator_optimizer_func(
            itertools.chain(
                self.discriminator_model_image_A.parameters(),
                self.discriminator_model_image_B.parameters(),
            )
        )
        discriminator_scheduler = discriminator_scheduler_func(discriminator_optimizer)

        return (
            [generator_optimizer, discriminator_optimizer],
            [
                generator_scheduler,
                discriminator_scheduler,
            ],
        )

    def save_cyclegan_output(
        self, image_A, image_B, fake_A_from_B, fake_B_from_A, current_stage
    ):
        # check if the log path exists, if not create one
        log_dir = Path(self.trainer.log_dir)
        log_dir.mkdir(parents=True, exist_ok=True)

        current_epoch = self.current_epoch

        if len(fake_A_from_B.shape) == 3 and fake_A_from_B.shape[0] == 3:
            out_fn = log_dir / f"{current_stage}_epoch_{current_epoch}_fake_A.png"
            save_rgb(out_fn, np.moveaxis(fake_A_from_B, 0, -1))

            out_fn = log_dir / f"{current_stage}_epoch_{current_epoch}_fake_B.png"
            save_rgb(out_fn, np.moveaxis(fake_B_from_A, 0, -1))

            out_fn = log_dir / f"{current_stage}_epoch_{current_epoch}_real_B.png"
            save_rgb(out_fn, np.moveaxis(image_B, 0, -1))

            out_fn = log_dir / f"{current_stage}_epoch_{current_epoch}_real_A.png"
            save_rgb(out_fn, np.moveaxis(image_A, 0, -1))
        else:
            out_fn = log_dir / f"{current_stage}_epoch_{current_epoch}_fake_A.tiff"
            imsave(out_fn, fake_A_from_B)

            out_fn = log_dir / f"{current_stage}_epoch_{current_epoch}_fake_B.tiff"
            imsave(out_fn, fake_B_from_A)

            out_fn = log_dir / f"{current_stage}_epoch_{current_epoch}_real_B.tiff"
            imsave(out_fn, image_B)

            out_fn = log_dir / f"{current_stage}_epoch_{current_epoch}_real_A.tiff"
            imsave(out_fn, image_A)

    # only for validation
    def run_step(self, batch, batch_idx):
        # get the data
        image_A = batch["IM"]
        image_B = batch["GT"]

        # run generators and discriminators
        fake_B_from_A = self.generator_model_image_AtoB(image_A)
        fake_A_from_B = self.generator_model_image_BtoA(image_B)

        predFakeA = self.discriminator_model_image_A(fake_A_from_B)
        predFakeB = self.discriminator_model_image_B(fake_B_from_A)

        pred_real = torch.ones_like(predFakeA, requires_grad=False)
        pred_fake = torch.zeros_like(predFakeA, requires_grad=False)

        # GAN loss
        gan_loss_B = self.gan_loss(predFakeB, pred_real)
        gan_loss_A = self.gan_loss(predFakeA, pred_real)
        gan_loss = (gan_loss_A + gan_loss_B) / 2

        # cycle loss
        cycle_loss_A = self.cycle_loss(
            self.generator_model_image_BtoA(fake_B_from_A), image_A
        )
        cycle_loss_B = self.cycle_loss(
            self.generator_model_image_AtoB(fake_A_from_B), image_B
        )
        cycle_loss = (cycle_loss_A + cycle_loss_B) / 2

        # identity loss
        fake_A_from_A = self.generator_model_image_BtoA(image_A)
        fake_B_from_B = self.generator_model_image_BtoA(image_B)
        identity_loss_A = self.identity_loss(fake_A_from_A, image_A)
        identity_loss_B = self.identity_loss(fake_B_from_B, image_B)
        identity_loss = (identity_loss_A + identity_loss_B) / 2

        # weighted sum into generator loss
        G_loss = (
            gan_loss * self.gan_w
            + cycle_loss * self.cycle_w
            + identity_loss * self.identity_w
        )

        # discriminator on domain A
        predRealA = self.discriminator_model_image_A(image_A)
        dis_loss_A_real = self.gan_loss(predRealA, pred_real)
        dis_loss_A_fake = self.gan_loss(predFakeA, pred_fake)

        # discriminator on domain B
        predRealB = self.discriminator_model_image_B(image_B)
        dis_loss_B_real = self.gan_loss(predRealB, pred_real)
        dis_loss_B_fake = self.gan_loss(predFakeB, pred_fake)

        # average over all dis loss
        D_loss = 0.5 * (
            dis_loss_A_fake + dis_loss_A_real + dis_loss_B_fake + dis_loss_B_real
        )

        output = OrderedDict({"generator_loss": G_loss, "discriminator_loss": D_loss})

        if self.verbose and batch_idx == 0:
            self.save_cyclegan_output(
                image_A[0].detach().cpu().numpy(),
                image_B[0].detach().cpu().numpy(),
                fake_A_from_B[0].detach().cpu().numpy(),
                fake_B_from_A[0].detach().cpu().numpy(),
                "val",
            )
        return output

    def training_step(self, batch):
        # get optimizer:
        g_opt, d_opt = self.optimizers()

        # get the batch
        image_A = batch["IM"]
        image_B = batch["GT"]

        # generate fake images
        fake_A_from_B = self.generator_model_image_BtoA(image_B)
        fake_B_from_A = self.generator_model_image_AtoB(image_A)

        ############################
        # train generators
        ############################
        self.toggle_optimizer(g_opt)
        # GAN loss
        predFakeA = self.discriminator_model_image_A(fake_A_from_B)
        predFakeB = self.discriminator_model_image_B(fake_B_from_A)

        pred_real = torch.ones_like(predFakeA, requires_grad=False)

        gan_loss_B = self.gan_loss(predFakeB, pred_real)
        gan_loss_A = self.gan_loss(predFakeA, pred_real)
        gan_loss = (gan_loss_A + gan_loss_B) / 2

        # cycle loss
        cycle_loss_A = self.cycle_loss(
            self.generator_model_image_BtoA(fake_B_from_A), image_A
        )
        cycle_loss_B = self.cycle_loss(
            self.generator_model_image_AtoB(fake_A_from_B), image_B
        )
        cycle_loss = (cycle_loss_A + cycle_loss_B) / 2

        # identity loss
        fake_A_from_A = self.generator_model_image_BtoA(image_A)
        fake_B_from_B = self.generator_model_image_BtoA(image_B)
        identity_loss_A = self.identity_loss(fake_A_from_A, image_A)
        identity_loss_B = self.identity_loss(fake_B_from_B, image_B)
        identity_loss = (identity_loss_A + identity_loss_B) / 2

        # weighted sum into generator loss
        g_loss = (
            gan_loss * self.gan_w
            + cycle_loss * self.cycle_w
            + identity_loss * self.identity_w
        )

        self.log(
            "train_g_loss",
            g_loss,
            on_step=True,
            on_epoch=True,
            prog_bar=True,
            logger=True,
            sync_dist=True,
        )
        # each individual loss only record in logger, not on progress bar
        self.log(
            "train_validity",
            gan_loss,
            on_step=True,
            on_epoch=True,
            logger=True,
            sync_dist=True,
        )
        self.log(
            "train_reconstruction",
            cycle_loss,
            on_step=True,
            on_epoch=True,
            logger=True,
            sync_dist=True,
        )
        self.log(
            "train_identity",
            identity_loss,
            on_step=True,
            on_epoch=True,
            logger=True,
            sync_dist=True,
        )

        g_opt.zero_grad()
        self.manual_backward(g_loss)
        g_opt.step()

        self.untoggle_optimizer(g_opt)

        ############################
        # train discriminators
        ############################
        self.toggle_optimizer(d_opt)

        fakeA_sample = self.fake_pool_A.push_and_pop(fake_A_from_B)
        fakeB_sample = self.fake_pool_B.push_and_pop(fake_B_from_A)
        predFakeA = self.discriminator_model_image_A(fakeA_sample)
        predFakeB = self.discriminator_model_image_B(fakeB_sample)

        pred_real = torch.ones_like(predFakeA, requires_grad=False)
        pred_fake = torch.zeros_like(predFakeA, requires_grad=False)

        # discriminator on domain A
        predRealA = self.discriminator_model_image_A(image_A)
        dis_loss_A_real = self.gan_loss(predRealA, pred_real)
        dis_loss_A_fake = self.gan_loss(predFakeA, pred_fake)

        # discriminator on domain B
        predRealB = self.discriminator_model_image_B(image_B)
        dis_loss_B_real = self.gan_loss(predRealB, pred_real)
        dis_loss_B_fake = self.gan_loss(predFakeB, pred_fake)

        # average over all dis loss
        d_loss = 0.5 * (
            dis_loss_A_fake + dis_loss_A_real + dis_loss_B_fake + dis_loss_B_real
        )

        self.log(
            "train_d_loss",
            d_loss,
            on_step=True,
            on_epoch=True,
            prog_bar=True,
            logger=True,
            sync_dist=True,
        )
        d_opt.zero_grad()
        self.manual_backward(d_loss)
        d_opt.step()

        self.untoggle_optimizer(d_opt)

        sch_g, sch_d = self.lr_schedulers()
        if self.trainer.is_last_batch:
            sch_g.step()
            sch_d.step()

            if self.verbose:
                self.save_cyclegan_output(
                    image_A[0].detach().cpu().numpy(),
                    image_B[0].detach().cpu().numpy(),
                    fake_A_from_B[0].detach().cpu().numpy(),
                    fake_B_from_A[0].detach().cpu().numpy(),
                    "train",
                )

    def validation_step(self, batch, batch_idx):
        val_loss = self.run_step(batch, batch_idx)

        self.log(
            "val_g_loss",
            val_loss["generator_loss"],
            on_step=True,
            on_epoch=True,
            prog_bar=True,
            logger=True,
            sync_dist=True,
        )
        self.log(
            "val_d_loss",
            val_loss["discriminator_loss"],
            on_step=True,
            on_epoch=True,
            prog_bar=True,
            logger=True,
            sync_dist=True,
        )



from pathlib import Path
from typing import Dict
from importlib import import_module
import numpy as np
from tifffile import imsave
from skimage.io import imsave as save_rgb
import lightning as pl
import torch
from mmv_im2im.utils.misc import parse_config_func
from mmv_im2im.utils.model_utils import init_weights, state_dict_simplification
from mmv_im2im.models.nets.gans import define_generator, define_discriminator


class Model(pl.LightningModule):
    def __init__(self, model_info_xx: Dict, train: bool = True, verbose: bool = False):
        super(Model, self).__init__()

        # Important: This property activates manual optimization.
        self.automatic_optimization = False

        self.verbose = verbose
        gen_init = None
        dis_init = None
        if "init_weight" in model_info_xx.net["generator"]:
            gen_init = model_info_xx.net["generator"].pop("init_weight")
        if "init_weight" in model_info_xx.net["discriminator"]:
            dis_init = model_info_xx.net["discriminator"].pop("init_weight")
        self.generator = define_generator(model_info_xx.net["generator"])
        self.discriminator = define_discriminator(model_info_xx.net["discriminator"])
        # self.sliding_window = model_info_xx["sliding_window_params"]
        if train:
            # initialize model weights
            if gen_init is not None:
                if Path(gen_init).is_file:
                    print(f"loading pre-train from {gen_init}")
                    try:
                        pre_train = torch.load(Path(gen_init))
                    except RuntimeError:
                        pre_train = torch.load(
                            Path(gen_init), map_location=torch.device("cpu")
                        )
                    try:
                        self.generator.load_state_dict(pre_train["state_dict"])
                    except Exception:
                        cleaned_state_dict = state_dict_simplification(
                            pre_train["state_dict"], cut="generator."
                        )
                        self.generator.load_state_dict(cleaned_state_dict, strict=False)
                else:
                    init_weights(self.generator, init_type=gen_init)
            else:
                init_weights(self.generator, init_type="normal")

            if dis_init is not None:
                if Path(dis_init).is_file:
                    try:
                        pre_train = torch.load(Path(dis_init))
                    except RuntimeError:
                        pre_train = torch.load(
                            Path(dis_init), map_location=torch.device("cpu")
                        )
                    self.discriminator.load_state_dict(pre_train["state_dict"])
                else:
                    init_weights(self.discriminator, init_type=dis_init)
            else:
                init_weights(self.discriminator, init_type="normal")

            # get loss functions
            loss_category = model_info_xx.criterion.pop("loss_type")
            loss_module = import_module("mmv_im2im.utils.gan_losses")
            loss_func = getattr(loss_module, loss_category)
            self.loss = loss_func(**model_info_xx.criterion)

            # get info of optimizer and scheduler
            self.optimizer_info = model_info_xx.optimizer
            self.scheduler_info = model_info_xx.scheduler

    def forward(self, x):
        return self.generator(x)

    def configure_optimizers(self):
        discriminator_optimizer_func = parse_config_func(
            self.optimizer_info["discriminator"]
        )
        discriminator_scheduler_func = parse_config_func(
            self.scheduler_info["discriminator"]
        )
        discriminator_optimizer = discriminator_optimizer_func(
            self.discriminator.parameters()
        )
        discriminator_scheduler = discriminator_scheduler_func(discriminator_optimizer)

        generator_optimizer_func = parse_config_func(self.optimizer_info["generator"])
        generator_scheduler_func = parse_config_func(self.scheduler_info["generator"])
        generator_optimizer = generator_optimizer_func(
            self.generator.parameters(),
        )
        generator_scheduler = generator_scheduler_func(generator_optimizer)

        return (
            [discriminator_optimizer, generator_optimizer],
            [
                discriminator_scheduler,
                generator_scheduler,
            ],
        )

    def save_pix2pix_output(self, image_A, image_B, fake_image, current_stage):
        # check if the log path exists, if not create one
        log_dir = Path(self.trainer.log_dir)
        log_dir.mkdir(parents=True, exist_ok=True)

        current_epoch = self.current_epoch

        if len(fake_image.shape) == 3 and fake_image.shape[0] == 3:
            out_fn = log_dir / f"{current_stage}_epoch_{current_epoch}_fake_B.png"
            save_rgb(out_fn, np.moveaxis(fake_image, 0, -1))

            out_fn = log_dir / f"{current_stage}_epoch_{current_epoch}_real_B.png"
            save_rgb(out_fn, np.moveaxis(image_B, 0, -1))

            out_fn = log_dir / f"{current_stage}_epoch_{current_epoch}_real_A.png"
            save_rgb(out_fn, np.moveaxis(image_A, 0, -1))
        else:
            out_fn = log_dir / f"{current_stage}_epoch_{current_epoch}_fake_B.tiff"
            imsave(out_fn, fake_image)

            out_fn = log_dir / f"{current_stage}_epoch_{current_epoch}_real_B.tiff"
            imsave(out_fn, image_B)

            out_fn = log_dir / f"{current_stage}_epoch_{current_epoch}_real_A.tiff"
            imsave(out_fn, image_A)

    def training_step(self, batch):
        # imageB : real image
        # imageA : condition image
        # conditional GAN refers generating a realistic image (image B as ground truth),
        # conditioned on image A
        image_A = batch["IM"]
        image_B = batch["GT"]

        # get optimizer:
        d_opt, g_opt = self.optimizers()

        #######################
        # calculate loss
        #######################
        # for generator
        self.toggle_optimizer(g_opt)
        fake_B = self.generator(image_A)
        g_loss = self.loss.generator_step(self.discriminator, image_A, image_B, fake_B)

        self.log(
            "train_g_loss",
            g_loss,
            on_step=True,
            on_epoch=True,
            prog_bar=True,
            logger=True,
            sync_dist=True,
        )
        g_opt.zero_grad()
        self.manual_backward(g_loss)
        g_opt.step()
        self.untoggle_optimizer(g_opt)

        # for discriminator
        self.toggle_optimizer(d_opt)
        fake_B = self.generator(image_A).detach()
        d_loss = self.loss.discriminator_step(
            self.discriminator, image_A, image_B, fake_B
        )

        self.log(
            "train_d_loss",
            d_loss,
            on_step=True,
            on_epoch=True,
            prog_bar=True,
            logger=True,
            sync_dist=True,
        )
        d_opt.zero_grad()
        self.manual_backward(d_loss)
        d_opt.step()
        self.untoggle_optimizer(d_opt)

        sch_d, sch_g = self.lr_schedulers()
        if self.trainer.is_last_batch:
            sch_g.step()
            sch_d.step()

    def validation_step(self, batch, batch_idx):
        image_A = batch["IM"]
        image_B = batch["GT"]

        fake_B = self.generator(image_A).detach()
        d_loss = self.loss.discriminator_step(
            self.discriminator, image_A, image_B, fake_B
        )
        self.log(
            "val_d_loss",
            d_loss,
            on_step=True,
            on_epoch=True,
            prog_bar=True,
            logger=True,
            sync_dist=True,
        )

        g_loss = self.loss.generator_step(self.discriminator, image_A, image_B, fake_B)
        self.log(
            "val_g_loss",
            g_loss,
            on_step=True,
            on_epoch=True,
            prog_bar=True,
            logger=True,
            sync_dist=True,
        )

        if self.verbose and batch_idx == 0:
            fake_images = fake_B.detach()
            fake_image = fake_images[0].cpu().numpy()

            image_A0 = image_A[0].detach().cpu().numpy()
            image_B0 = image_B[0].detach().cpu().numpy()

            self.save_pix2pix_output(image_A0, image_B0, fake_image, "val")



# The following file was adapted from https://github.com/AllenCellModeling/pytorch_fnet

import torch


class Net(torch.nn.Module):
    def __init__(self, depth=4, mult_chan=32, in_channels=1, out_channels=1):
        super().__init__()
        self.depth = depth
        self.mult_chan = mult_chan
        self.in_channels = in_channels
        self.out_channels = out_channels

        self.net_recurse = _Net_recurse(
            n_in_channels=self.in_channels,
            mult_chan=self.mult_chan,
            depth_parent=self.depth,
            depth=self.depth,
        )
        self.conv_out = torch.nn.Conv3d(
            self.mult_chan, self.out_channels, kernel_size=3, padding=1
        )

    def forward(self, x):
        x_rec = self.net_recurse(x)
        return self.conv_out(x_rec)


class _Net_recurse(torch.nn.Module):
    def __init__(self, n_in_channels, mult_chan=2, depth_parent=0, depth=0):
        """Class for recursive definition of U-network.p
        Parameters
        ----------
        in_channels
            Number of channels for input.
        mult_chan
            Factor to determine number of output channels
        depth
            If 0, this subnet will only be convolutions that double the channel
            count.
        """
        super().__init__()

        self.depth = depth

        if self.depth == depth_parent:
            n_out_channels = mult_chan
        else:
            n_out_channels = n_in_channels * mult_chan

        self.sub_2conv_more = SubNet2Conv(n_in_channels, n_out_channels)
        if depth > 0:
            self.sub_2conv_less = SubNet2Conv(2 * n_out_channels, n_out_channels)
            self.conv_down = torch.nn.Conv3d(
                n_out_channels, n_out_channels, 2, stride=2
            )
            self.bn0 = torch.nn.BatchNorm3d(n_out_channels)
            self.relu0 = torch.nn.ReLU()
            self.convt = torch.nn.ConvTranspose3d(
                2 * n_out_channels, n_out_channels, kernel_size=2, stride=2
            )
            self.bn1 = torch.nn.BatchNorm3d(n_out_channels)
            self.relu1 = torch.nn.ReLU()
            self.sub_u = _Net_recurse(
                n_out_channels,
                mult_chan=2,
                depth_parent=depth_parent,
                depth=(depth - 1),
            )

    def forward(self, x):
        if self.depth == 0:
            return self.sub_2conv_more(x)
        else:  # depth > 0
            x_2conv_more = self.sub_2conv_more(x)
            x_conv_down = self.conv_down(x_2conv_more)
            x_bn0 = self.bn0(x_conv_down)
            x_relu0 = self.relu0(x_bn0)
            x_sub_u = self.sub_u(x_relu0)
            x_convt = self.convt(x_sub_u)
            x_bn1 = self.bn1(x_convt)
            x_relu1 = self.relu1(x_bn1)
            x_cat = torch.cat((x_2conv_more, x_relu1), 1)  # concatenate
            x_2conv_less = self.sub_2conv_less(x_cat)
        return x_2conv_less


class SubNet2Conv(torch.nn.Module):
    def __init__(self, n_in, n_out):
        super().__init__()
        self.conv1 = torch.nn.Conv3d(n_in, n_out, kernel_size=3, padding=1)
        self.bn1 = torch.nn.BatchNorm3d(n_out)
        self.relu1 = torch.nn.ReLU()
        self.conv2 = torch.nn.Conv3d(n_out, n_out, kernel_size=3, padding=1)
        self.bn2 = torch.nn.BatchNorm3d(n_out)
        self.relu2 = torch.nn.ReLU()

    def forward(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu1(x)
        x = self.conv2(x)
        x = self.bn2(x)
        x = self.relu2(x)
        return x



from typing import Optional
import numpy as np


def simplified_instance_IoU(
    mask: np.ndarray, pred: np.ndarray, exclusion_mask: Optional[np.ndarray] = None
):
    if exclusion_mask is not None:
        exclusion_mask = np.squeeze(exclusion_mask)
        assert (
            exclusion_mask.shape == mask.shape
        ), "exclustion mask and gt have different sizes"
        mask[exclusion_mask > 0] = 0
        pred[exclusion_mask > 0] = 0

    # clean up ground truth by removing all objects touching boundary
    # also, boudary is defined by 5 pixels within the actual border
    boundary_template = np.zeros_like(mask)
    if len(mask.shape) == 3:
        boundary_template[:, :5, :] = 1
        boundary_template[:, -5:, :] = 1
        boundary_template[:, :, :5] = 1
        boundary_template[:, :, -5:] = 1
    elif len(mask.shape) == 2:
        boundary_template[:5, :] = 1
        boundary_template[-5:, :] = 1
        boundary_template[:, :5] = 1
        boundary_template[:, -5:] = 1
    else:
        raise ValueError("bad image size in IoU")

    bd_idx = list(np.unique(mask[boundary_template > 0]))
    adjusted_mask = mask.copy()
    for idx in bd_idx:
        adjusted_mask[mask == idx] = 0

    # loop through all instances in GT and compare with pred
    valid_idx = list(np.unique(adjusted_mask[adjusted_mask > 0]))
    iou_list = []
    for gt_idx in valid_idx:
        gt_obj = adjusted_mask == gt_idx

        # find corresponding prediction mask
        eval_overlap = pred.copy()
        eval_overlap[gt_obj == 0] = 0
        pred_candidates = list(np.unique(eval_overlap[eval_overlap > 0]))
        if len(pred_candidates) == 0:
            continue
        pred_match = []
        for pred_idx in pred_candidates:
            overlap_size = np.count_nonzero(eval_overlap == pred_idx)
            predict_size = np.count_nonzero(pred == pred_idx) + 0.0000001
            if overlap_size / predict_size > 0.25:
                pred_match.append(pred_idx)
        if len(pred_match) == 0:
            continue
        pred_mask = np.zeros_like(mask)
        for pred_idx in pred_match:
            pred_mask[pred == pred_idx] = 1

        # calculate IOU
        i_size = np.count_nonzero(np.logical_and(pred_mask > 0, gt_obj > 0))
        u_size = np.count_nonzero(np.logical_or(pred_mask > 0, gt_obj > 0))
        score = i_size / (u_size + 0.000001)
        iou_list.append(score)

    if len(iou_list) == 0:
        return 0
    else:
        return sum(iou_list) / len(iou_list)



########################################################
# ####       general data module for          ########
# ####   paired and unpaired 3D/2D images         #######
# ###  (mostly for FCN or CGAN-like models)     ######
#
#
# About data in a batch:
# We woudl expect 3 parts, image_source, image_target,
# and image_cmap (cmap: cost map), where image_cmap
# can be optional. Note that image_target could be masks
# (e.g. for segmentation) or images (e.g. for labelfree)
########################################################
import random
from torch.utils.data import random_split
from monai.data import DataLoader
import lightning as pl
from mmv_im2im.utils.for_transform import parse_monai_ops  # , custom_preproc_to_tio
from mmv_im2im.utils.misc import (
    generate_dataset_dict_monai,
    parse_config_func_without_params,
)
import monai
from monai.data import list_data_collate


class Im2ImDataModule(pl.LightningDataModule):
    def __init__(self, data_cfg):
        super().__init__()
        self.data_path = data_cfg.data_path
        self.category = data_cfg.category

        # train/val split
        self.train_val_ratio = data_cfg.dataloader.train_val_ratio or 0.2
        self.train_set = None
        self.val_set = None

        if data_cfg.preprocess is None:
            self.preproc = None
        else:
            self.preproc = parse_monai_ops(data_cfg.preprocess)

        if data_cfg.augmentation is None:
            self.augment = None
        else:
            self.augment = parse_monai_ops(data_cfg.augmentation)

        if self.preproc is None and self.augment is not None:
            self.transform = self.augment
        elif self.preproc is not None and self.augment is None:
            self.transform = self.preproc
        elif self.preproc is None and self.augment is None:
            self.transform = None
        else:
            self.transform = monai.transforms.Compose([self.preproc, self.augment])

        # parameters for dataloader
        self.dataloader_info = data_cfg.dataloader

    def prepare_data(self):
        pass

    def setup(self, stage=None):
        dataset_list = generate_dataset_dict_monai(self.data_path)

        if self.category == "unpair":
            if "train" in dataset_list and "val" in dataset_list:
                raise ValueError(
                    "currently, unpaired data loading does not support "
                    "pre-partitioned dataset"
                )
            shuffled_dataset_list = dataset_list.copy()
            random.shuffle(shuffled_dataset_list)
            for i, shuffled_ds in enumerate(shuffled_dataset_list):
                dataset_list[i]["GT"] = shuffled_ds["GT"]

        self.data = dataset_list

        if "train" in self.data and "val" in self.data:
            train_subjects = self.data["train"]
            val_subjects = self.data["val"]
        else:
            num_subjects = len(self.data)
            num_val_subjects = int(round(num_subjects * self.train_val_ratio))
            num_train_subjects = num_subjects - num_val_subjects
            splits = num_train_subjects, num_val_subjects
            train_subjects, val_subjects = random_split(self.data, splits)
        assert len(val_subjects) > 0
        assert len(train_subjects) > 0
        self.val_data = val_subjects
        self.train_data = train_subjects

    def train_dataloader(self):
        train_loader_info = self.dataloader_info.train
        train_dataset_func = parse_config_func_without_params(
            train_loader_info.dataloader_type
        )
        # gather the data filepath into dataset
        train_data = self.train_data
        if train_loader_info.partial_loader < 1.0:
            num_load = int(train_loader_info.partial_loader * len(self.train_data))
            from sklearn.utils import shuffle

            train_data = shuffle(train_data)
            train_data = train_data[:num_load]
        if len(train_loader_info.dataset_params) > 0:
            train_dataset = train_dataset_func(
                data=train_data,
                transform=self.transform,
                **train_loader_info.dataset_params,
            )
        else:
            train_dataset = train_dataset_func(
                data=train_data, transform=self.transform
            )
        # wrap the dataset into dataloader
        train_dataloader = DataLoader(
            train_dataset,
            shuffle=True,
            collate_fn=list_data_collate,
            **train_loader_info.dataloader_params,
        )
        return train_dataloader

    def val_dataloader(self):
        val_loader_info = self.dataloader_info.val
        val_dataset_func = parse_config_func_without_params(
            val_loader_info.dataloader_type
        )
        if len(val_loader_info.dataset_params) > 0:
            val_dataset = val_dataset_func(
                data=self.val_data,
                transform=self.preproc,
                **val_loader_info.dataset_params,
            )
        else:
            val_dataset = val_dataset_func(data=self.val_data, transform=self.preproc)
        val_dataloader = DataLoader(
            val_dataset,
            shuffle=False,
            collate_fn=list_data_collate,
            **val_loader_info.dataloader_params,
        )
        return val_dataloader



from typing import List, Dict
from functools import partial
from mmv_im2im.utils.misc import parse_config, parse_config_func_without_params
from monai.transforms import Compose, Lambdad, Lambda
import inspect


def center_crop(img, target_shape):
    # target_shape is smaller than img.shape
    target_x = target_shape[-1]
    img_x = img.shape[-1]
    diff_x = img_x - target_x
    half_x = diff_x // 2

    target_y = target_shape[-2]
    img_y = img.shape[-2]
    diff_y = img_y - target_y
    half_y = diff_y // 2

    if len(target_shape) == 3:
        target_z = target_shape[-3]
        img_z = img.shape[-3]
        diff_z = img_z - target_z
        half_z = diff_z // 2

        return img[
            half_z : -(diff_z - half_z),
            half_y : -(diff_y - half_y),
            half_x : -(diff_x - half_x),
        ]
    else:
        return img[half_y : -(diff_y - half_y), half_x : -(diff_x - half_x)]


def parse_monai_ops(trans_func: List[Dict]):
    # Here, we will use the Compose function in MONAI to merge
    # all transformations. If any trnasformation not from MONAI,
    # a MONAI Lambda function will be used to wrap around it.
    trans_list = []

    # loop throught the config
    for func_info in trans_func:
        if func_info["module_name"] == "monai.transforms":
            if func_info["func_name"] == "LoadImaged":
                # Here, we handle the LoadImaged seperatedly to allow bio-reader
                from mmv_im2im.utils.misc import monai_bio_reader
                from monai.transforms import LoadImaged

                trans_list.append(
                    LoadImaged(reader=monai_bio_reader, **func_info["params"])
                )
            else:
                trans_list.append(parse_config(func_info))
        else:
            my_func = parse_config_func_without_params(func_info)
            func_params = func_info["params"]
            apply_keys = func_params.pop("keys")

            # check if any other params
            if len(func_params) > 0:
                if inspect.isclass(my_func):
                    callable_func = my_func(**func_params)
                else:
                    callable_func = partial(my_func, **func_params)
            else:
                callable_func = my_func

            trans_list.append(Lambdad(keys=apply_keys, func=callable_func))

    return Compose(trans_list)


def parse_monai_ops_vanilla(trans_func: List[Dict]):
    # Here, we will use the Compose function in MONAI to merge
    # all transformations. If any trnasformation not from MONAI,
    # a MONAI Lambda function will be used to wrap around it.
    trans_list = []
    # loop throught the config
    for func_info in trans_func:
        if func_info["module_name"] == "monai.transforms":
            trans_list.append(parse_config(func_info))
        else:
            my_func = parse_config_func_without_params(func_info)
            if "params" in func_info:
                func_params = func_info["params"]
                callable_func = partial(my_func, **func_params)
            else:
                callable_func = my_func
            trans_list.append(Lambda(func=callable_func))

    return Compose(trans_list)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This sample script will get deployed in the bin directory of the
users' virtualenv when the parent module is installed using pip.
"""
import logging
import sys
import traceback
import tempfile
from pathlib import Path

# import torch


from mmv_im2im import ProjectTester, ProjectTrainer
from mmv_im2im.map_extractor import MapExtractor
from mmv_im2im.configs.config_base import (
    ProgramConfig,
    parse_adaptor,
    configuration_validation,
)


###############################################################################

log = logging.getLogger()
logging.basicConfig(
    level=logging.INFO,
    format="[%(levelname)4s:%(lineno)4s %(asctime)s] %(message)s",  # noqa E501
)

###############################################################################
TRAIN_MODE = "train"
INFER_MODE = "inference"
MAP_MODE = "uncertainty_map"

###############################################################################


def main():
    cfg = parse_adaptor(config_class=ProgramConfig)

    # check the configurations to make sure no conflicting settings
    cfg = configuration_validation(cfg)

    try:
        # check gpu option
        # assert torch.cuda.is_available(), "GPU is not available."
        # torch.cuda.set_device(torch.device("cuda:0"))

        if cfg.mode.lower() == TRAIN_MODE:
            if (
                cfg.data.dataloader.train.dataloader_type["func_name"]
                == "PersistentDataset"
            ):
                # currently, train and val have to use persistent load together, or
                # both not use and their root cache dirs are the same. See the
                # validation part in config
                cache_root = Path(cfg.data.dataloader.train.dataset_params["cache_dir"])
                cache_root.mkdir(exist_ok=True)
                with tempfile.TemporaryDirectory(dir=cache_root) as tmp_exp:
                    train_cache = Path(tmp_exp) / "train"
                    val_cache = Path(tmp_exp) / "val"
                    cfg.data.dataloader.train.dataset_params["cache_dir"] = train_cache
                    cfg.data.dataloader.val.dataset_params["cache_dir"] = val_cache
                    exe = ProjectTrainer(cfg)
                    exe.run_training()
            else:
                exe = ProjectTrainer(cfg)
                exe.run_training()
        elif cfg.mode.lower() == INFER_MODE:
            exe = ProjectTester(cfg)
            exe.run_inference()
        elif cfg.mode.lower() == MAP_MODE:
            exe = MapExtractor(cfg)
            exe.run_inference()
        else:
            log.error(f"Mode {cfg.mode} is not supported yet")
            sys.exit(1)

    except Exception as e:
        log.error("=============================================")
        log.error("\n\n" + traceback.format_exc())
        log.error("=============================================")
        log.error("\n\n" + str(e) + "\n")
        log.error("=============================================")
        sys.exit(1)


###############################################################################
# Allow caller to directly run this module (usually in development scenarios)

if __name__ == "__main__":
    main()



import torch
import random


# adapted from
# https://github.com/eriklindernoren/PyTorch-GAN/blob/master/implementations/cyclegan/utils.py#L13
class ReplayBuffer:
    def __init__(self, max_size=50):
        assert (
            max_size > 0
        ), "Empty buffer or trying to create a black hole. Be careful."
        self.max_size = max_size
        self.data = []

    def push_and_pop(self, data):
        to_return = []
        for element in data.data:
            element = torch.unsqueeze(element, 0)
            if len(self.data) < self.max_size:
                self.data.append(element)
                to_return.append(element)
            else:
                if random.uniform(0, 1) > 0.5:
                    i = random.randint(0, self.max_size - 1)
                    to_return.append(self.data[i].clone())
                    self.data[i] = element
                else:
                    to_return.append(element)
        return torch.cat(to_return)



import numpy as np
from typing import Union
from numba import jit
from scipy.ndimage.measurements import find_objects
from scipy.ndimage.morphology import binary_fill_holes
from bioio.writers import OmeTiffWriter
from bioio import BioImage
from tqdm import tqdm
from pathlib import Path
import warnings
from torch import from_numpy
from monai.data.meta_tensor import MetaTensor

from mmv_im2im.utils.misc import generate_dataset_dict


@jit(nopython=True)
def pairwise_python(X):
    M = X.shape[0]
    N = X.shape[1]
    D = np.empty((M, M), dtype=np.float32)
    for i in range(M):
        for j in range(M):
            d = 0.0
            for k in range(N):
                tmp = X[i, k] - X[j, k]
                d += tmp * tmp
            D[i, j] = np.sqrt(d)
    return D


def fill_label_holes(lbl_img, **kwargs):
    """Fill small holes in label image."""

    def grow(sl, interior):
        return tuple(
            slice(s.start - int(w[0]), s.stop + int(w[1])) for s, w in zip(sl, interior)
        )

    def shrink(interior):
        return tuple(slice(int(w[0]), (-1 if w[1] else None)) for w in interior)

    objects = find_objects(lbl_img)
    lbl_img_filled = np.zeros_like(lbl_img)
    for i, sl in enumerate(objects, 1):
        if sl is None:
            continue
        interior = [(s.start > 0, s.stop < sz) for s, sz in zip(sl, lbl_img.shape)]
        shrink_slice = shrink(interior)
        grown_mask = lbl_img[grow(sl, interior)] == i
        mask_filled = binary_fill_holes(grown_mask, **kwargs)[shrink_slice]
        lbl_img_filled[sl][mask_filled] = i
    return lbl_img_filled


def generate_center_image_2d(instance, center, ids):
    """
    Generates a `center_image` which is one (True) for all center locations and
    zero (False) otherwise.

    Parameters
    ----------
    instance: numpy array
        `instance` image containing unique `ids` for each object (YX)
         or present in a one-hot encoded style where each object is one in it
         own slice and zero elsewhere.
    center: string
        One of 'centroid', 'approximate-medoid' or 'medoid'.
    ids: list
        Unique ids corresponding to the objects present in the instance image.
    one_hot: boolean
        True (in this case, `instance` has shape DYX) or False (in this case,
        `instance` has shape YX).
    """

    center_image = np.zeros(instance.shape, dtype=bool)
    for j, id in enumerate(ids):
        y, x = np.where(instance == id)
        if len(y) != 0 and len(x) != 0:
            if center == "centroid":
                ym, xm = np.mean(y), np.mean(x)
            elif center == "approximate-medoid":
                ym_temp, xm_temp = np.median(y), np.median(x)
                imin = np.argmin((x - xm_temp) ** 2 + (y - ym_temp) ** 2)
                ym, xm = y[imin], x[imin]
            elif center == "medoid":
                dist_matrix = pairwise_python(np.vstack((x, y)).transpose())
                imin = np.argmin(np.sum(dist_matrix, axis=0))
                ym, xm = y[imin], x[imin]
            else:
                raise NotImplementedError("error in center method")
            center_image[int(np.round(ym)), int(np.round(xm))] = True
    return center_image


def generate_center_image_3d(instance, center, ids, anisotropy_factor, speed_up):
    center_image = np.zeros(instance.shape, dtype=bool)
    instance_downsampled = instance[
        :, :: int(speed_up), :: int(speed_up)
    ]  # down sample in x and y
    for j, id in enumerate(ids):
        z, y, x = np.where(instance_downsampled == id)
        if len(y) != 0 and len(x) != 0:
            if center == "centroid":
                zm, ym, xm = np.mean(z), np.mean(y), np.mean(x)
            elif center == "approximate-medoid":
                zm_temp, ym_temp, xm_temp = np.median(z), np.median(y), np.median(x)
                imin = np.argmin(
                    (x - xm_temp) ** 2
                    + (y - ym_temp) ** 2
                    + (anisotropy_factor * (z - zm_temp)) ** 2
                )
                zm, ym, xm = z[imin], y[imin], x[imin]
            elif center == "medoid":
                dist_matrix = pairwise_python(
                    np.vstack(
                        (speed_up * x, speed_up * y, anisotropy_factor * z)
                    ).transpose()
                )
                imin = np.argmin(np.sum(dist_matrix, axis=0))
                zm, ym, xm = z[imin], y[imin], x[imin]
            center_image[
                int(np.round(zm)),
                int(np.round(speed_up * ym)),
                int(np.round(speed_up * xm)),
            ] = True
    return center_image


def generate_center_image(instance, center, ids, anisotropy_factor=1, speed_up=1):
    if len(instance.shape) == 3:
        return generate_center_image_3d(
            instance, center, ids, anisotropy_factor, speed_up
        )
    elif len(instance.shape) == 2:
        return generate_center_image_2d(
            instance,
            center,
            ids,
        )
    else:
        raise ValueError("instance image must be either 2D or 3D")


def prepare_embedseg_cache(
    data_path: Union[str, Path], cache_path: Union[str, Path], data_cfg
):
    data_path = Path(data_path)
    cache_path = Path(cache_path)
    dataset_list = generate_dataset_dict(data_path)
    patch_size = None

    # parse the method for centroid computation
    if data_cfg.extra is not None:
        if "center_method" in data_cfg.extra:
            center_method = data_cfg.extra["center_method"]
        if "patch_size" in data_cfg.extra:
            patch_size = data_cfg.extra["patch_size"]

    # get some basic statistics from the data and do some basic validation
    min_xy = 65535
    min_z = 65535
    spatial_dim = 2
    for ds in dataset_list:
        fn = ds["source_fn"]
        reader = BioImage(fn)
        this_minXY = min(reader.dims.X, reader.dims.Y)
        min_xy = min((this_minXY, min_xy))
        if reader.dims.Z > 1 and spatial_dim == 2:
            spatial_dim = 3
        min_z = min((reader.dims.Z, min_z))
        assert this_minXY >= 128, "{fn}: XY dimension smaller than 128, not good"
        if spatial_dim == 3:
            assert reader.dims.Z >= 16, "{fn} has less than 16 Z slices, not good"

    if patch_size is None:
        crop_size = 128 * (min_xy // 128)
        if spatial_dim == 3:
            crop_size_z = min((32, min_z))
            new_patch_size = [crop_size_z, crop_size, crop_size]
        else:
            new_patch_size = [crop_size, crop_size]
        warnings.warn(
            UserWarning(
                f"A patch_size is determined from data. MAKE SURE to set data.patch_size as {new_patch_size}."  # noqa E501
            )
        )

    else:
        spatial_dim = len(patch_size)
        if spatial_dim == 3:
            crop_size_z = patch_size[0]
            crop_size = min(patch_size[1:])
        else:
            crop_size = min(patch_size)

    if spatial_dim == 3:
        reader_params = {"dimension_order_out": "ZYX", "C": 0, "T": 0}
        raw_reader_params = {"dimension_order_out": "CZYX", "T": 0}
    else:
        reader_params = {"dimension_order_out": "YX", "C": 0, "T": 0, "Z": 0}
        raw_reader_params = {"dimension_order_out": "CYX", "T": 0, "Z": 0}

    # loop through the dataset
    for ds in tqdm(dataset_list):
        # get instance segmentation labels
        instance_reader = BioImage(ds["target_fn"])
        instance = instance_reader.get_image_data(**reader_params)

        # get raw image
        image_reader = BioImage(ds["source_fn"])
        image = image_reader.get_image_data(**raw_reader_params)

        # check if costmap exists
        fn_base = ds["source_fn"].stem[:-3]
        cm_fn = ds["source_fn"].parent / f"{fn_base}_CM.tiff"
        costmap_flag = False
        if cm_fn.is_file():
            costmap_flag = True
            cm_reader = BioImage(cm_fn)
            costmap = cm_reader.get_image_data(**reader_params)

        # parse filename
        fn_base = Path(ds["source_fn"]).stem[:-3]  # get rid of "_IM"

        # get all obejcts
        instance_np = np.array(instance, copy=False)
        object_mask = instance_np > 0

        ids = np.unique(instance_np[object_mask])
        ids = ids[ids != 0]

        # loop over instances
        for j, id in enumerate(ids):
            if spatial_dim == 2:
                h, w = instance.shape
                y, x = np.where(instance_np == id)
                ym, xm = np.mean(y), np.mean(x)

                jj = int(np.clip(ym - crop_size / 2, 0, h - crop_size))
                ii = int(np.clip(xm - crop_size / 2, 0, w - crop_size))

                # only crop patches away from image borders
                if instance[jj : jj + crop_size, ii : ii + crop_size].shape == (
                    crop_size,
                    crop_size,
                ):
                    im_crop = image[:, jj : jj + crop_size, ii : ii + crop_size]
                    instance_crop = instance[jj : jj + crop_size, ii : ii + crop_size]
                    center_image_crop = generate_center_image(
                        instance_crop, center_method, ids
                    )
                    class_image_crop = object_mask[
                        jj : jj + crop_size, ii : ii + crop_size
                    ]
                    if costmap_flag:
                        costmap_crop = costmap[jj : jj + crop_size, ii : ii + crop_size]
                    dim_order = "YX"
                else:
                    continue

            elif spatial_dim == 3:
                d, h, w = instance.shape
                z, y, x = np.where(instance_np == id)
                zm, ym, xm = np.mean(z), np.mean(y), np.mean(x)
                kk = int(np.clip(zm - crop_size_z / 2, 0, d - crop_size_z))
                jj = int(np.clip(ym - crop_size / 2, 0, h - crop_size))
                ii = int(np.clip(xm - crop_size / 2, 0, w - crop_size))

                if instance[
                    kk : kk + crop_size_z, jj : jj + crop_size, ii : ii + crop_size
                ].shape == (crop_size_z, crop_size, crop_size):
                    im_crop = image[
                        :,
                        kk : kk + crop_size_z,
                        jj : jj + crop_size,
                        ii : ii + crop_size,
                    ]
                    instance_crop = instance[
                        kk : kk + crop_size_z,
                        jj : jj + crop_size,
                        ii : ii + crop_size,
                    ]
                    center_image_crop = generate_center_image(
                        instance_crop,
                        center_method,
                        ids,
                        anisotropy_factor=1,
                        speed_up=1,
                    )
                    class_image_crop = object_mask[
                        kk : kk + crop_size_z,
                        jj : jj + crop_size,
                        ii : ii + crop_size,
                    ]
                    if costmap_flag:
                        costmap_crop = costmap[
                            kk : kk + crop_size_z,
                            jj : jj + crop_size,
                            ii : ii + crop_size,
                        ]
                    dim_order = "ZYX"
                else:
                    continue

            else:
                raise ValueError(
                    "error in spatial dimension when preparing embedseg dataset"
                )

            if im_crop.shape[0] == 1:
                OmeTiffWriter.save(
                    im_crop[0,],
                    cache_path / f"{fn_base}_{j:04d}_IM.tiff",
                    dim_order=dim_order,
                )
            else:
                OmeTiffWriter.save(
                    im_crop,
                    cache_path / f"{fn_base}_{j:04d}_IM.tiff",
                    dim_order="C" + dim_order,
                )
            OmeTiffWriter.save(
                instance_crop.astype(np.uint16),
                cache_path / f"{fn_base}_{j:04d}_GT.tiff",
                dim_order=dim_order,
            )
            OmeTiffWriter.save(
                center_image_crop.astype(np.uint8),
                cache_path / f"{fn_base}_{j:04d}_CE.tiff",
                dim_order=dim_order,
            )
            OmeTiffWriter.save(
                class_image_crop.astype(np.uint8),
                cache_path / f"{fn_base}_{j:04d}_CL.tiff",
                dim_order=dim_order,
            )
            if costmap_flag:
                OmeTiffWriter.save(
                    costmap_crop.astype(float),
                    cache_path / f"{fn_base}_{j:04d}_CM.tiff",
                    dim_order=dim_order,
                )


def prepare_embedseg_tensor(
    instance_batch: MetaTensor,
    spatial_dim: int,
    center_method: str = "centroid",
):
    """
    Parameters:
    ------------
        instance: instance segmentation masks of shape BYX or BZYX
        spatial_dim: 2 or 3
        crop_size: values from cropping of shape YX or ZYX

    Return:
    ----------
        class_labels: MetaTensor
        center_images: MetaTensor
    """

    assert instance_batch.is_batch, "only batch is supported currently"
    num_samples = instance_batch.size()[0]

    center_images_list = []
    class_labels_list = []
    for sample_idx in range(num_samples):
        instance = instance_batch[sample_idx].detach().cpu().numpy()

        assert instance.shape[0] == 1, "ground truth has more than 1 channel"
        instance = np.squeeze(instance, axis=0)
        instance_np = np.array(instance, copy=False)

        class_image = instance_np > 0
        ids = np.unique(instance_np[class_image])
        ids = ids[ids != 0]
        center_image = generate_center_image(instance, center_method, ids)

        class_image = np.expand_dims(class_image, axis=0)
        class_labels_list.append(class_image.astype(np.ubyte))
        center_image = np.expand_dims(center_image, axis=0)
        center_images_list.append(center_image.astype(bool))

    # stack into batch and covert to MetaTensor
    center_images = from_numpy(np.stack(center_images_list, axis=0))
    class_labels = from_numpy(np.stack(class_labels_list, axis=0))

    center_images = MetaTensor(center_images)
    class_labels = MetaTensor(class_labels)

    # move to proer device
    if instance_batch.is_cuda:
        current_device = instance_batch.get_device()
        center_images = center_images.to(current_device)
        class_labels = class_labels.to(current_device)

    return class_labels, center_images
